from __future__ import annotations
from typing import Tuple, Any, Mapping, Union, List
import warnings
from abc import abstractmethod, ABC

import nidaqmx
import nidaqmx.stream_readers
import nidaqmx.stream_writers
import numpy as np
from frozendict import frozendict

from ..exceptions import SingletonError
from ..configs import Config, HoistedConfig
from ..data import RingBuffer, TimestampedRingBuffer
from ..tools import HAS
from ..registries import CallbackRegistry, Provider
from ..requests import CallbackRequest


"""
Components and sub-components for National Instruments Data Acquisition Devices.
"""


"""
Abstract Components
"""


@Provider()
class StreamingComponent:
    name = "Streaming Component"

    def __init__(self,
                 config: Config):
        self.device_id = config.device_id
        ################################################################################################################
        # PRE-ALLOCATE
        ################################################################################################################
        # pre-allocated
        #: int: buffer size for each channel
        self.buffer_size = None
        #: np.dtype: dtype for collected data
        self.data_type = None
        #: np.ndarray: buffer object of size channels x buffer size
        self.data = None

        ################################################################################################################
        # INITIALIZATION FROM CONFIG
        ################################################################################################################
        #: int: sampling rate for all DAQ streams (Hz)
        self.sampling_frequency = config.sampling_frequency
        #: int: buffer rate for daq
        self.buffer_frequency = config.buffer_frequency
        #: float: timeout for DAQ interaction (s)
        self.timeout = config.timeout
        #: frozendict: channel index hashmap
        self.channel_index = frozendict()
        #: frozendict: dictionary of individual channel strings
        self.channel_ids = frozendict()
        #: int: number of channels
        self.num_channels = 0
        #: str: flattened channel string for DAQ
        self.channel_str = ""
        #: frozendict: channel - buffer row mapping
        self.buffer_index = frozendict()
        # ingest channel information
        self.retrieve_channel_info(config)
        self.output_file = (
            config.save_location.joinpath(config.file_header + self.name.lower().replace(" ", "_")))
        self.ring_buffer = None
        #: nidaqmx.Task: streaming object & is really only called in subclass's init
        self.streamer = None  # self.generate_streamer
        # #: nidaqmx.stream_readers or nidaqmx.stream_writers: stream processor object, called subclass init
        self.stream_processor = None  # self.generate_processor

    def __str__(self):
        return self.name

#    def find_request_callbacks(self) -> StreamingComponent:
#        self.request_callbacks = [self.callbacks.get(key) for key in self.callbacks if "request" in key]
#
#    def find_update_callbacks(self) -> StreamingComponent:
#        self.update_callbacks = [(self.callbacks.get("".join([key, "_update"])), value)
#                                 for key, value in self.buffer_index.items()]

    @abstractmethod
    def generate_processor(self) -> StreamingComponent:
        ...

    @abstractmethod
    def generate_streamer(self) -> StreamingComponent:
        ...

    @abstractmethod
    def generate_ni_buffer(self, config: "Config"):
        ...

    def generate_ring_buffer(self) -> StreamingComponent:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.ring_buffer = TimestampedRingBuffer(element_depth=self.num_channels,
                                                     element_length=self.buffer_size,
                                                     ring_size=self.buffer_frequency * 10,
                                                     path=self.output_file,
                                                     dtype=self.data_type)

    @abstractmethod
    def retrieve_channel_info(self, config: Config) -> StreamingComponent:
        """
        Retrieves dictionary of channel index hashmap, channel string ids, number of channels,
        and flattened channel string
        """
        pass

    @abstractmethod
    def start_component(self) -> StreamingComponent:
        ...

    def stream_request(self) -> int:
        """
        Request in-place update of a data stream

        :return: 0
        :rtype: int
        """
        # READ DATA
        self._stream_callback_read()
        # UPDATE RING BUFFER
        self.ring_buffer.add_data(self.data)
        # USE NEW DATA FOR CALLBACKS
        self._stream_callback_calls(self.ring_buffer.get_data())
        return 0

    @abstractmethod
    def _stream_callback_read(self):
        ...

    @abstractmethod
    def _stream_callback_calls(self, data):
        ...

    def _stream_callback(self, task_handle: str, every_n_samples_event_type: Any, number_of_samples: int,
                         callback_data: np.ndarray) -> int:
        """
        In place update of a data stream

        :param task_handle: task handle for nidaqmx task
        :type task_handle: str
        :param every_n_samples_event_type: specifies the event is an every_n_samples_event
        :type every_n_samples_event_type: Any
        :param number_of_samples: samples to collect before callback
        :type number_of_samples: int
        :param callback_data: reference to the data buffer
        :type callback_data: np.ndarray
        :return: 0
        :rtype: int
        """
        # READ DATA
        self._stream_callback_read()
        # UPDATE RING BUFFER
        self.ring_buffer.add_data(self.data)
        # USE NEW DATA FOR CALLBACKS
        self._stream_callback_calls(self.ring_buffer.get_data())
        return 0

    def destroy(self):
        try:
            # Make sure it's actually a defined task and not the base component or empty streamer
            assert self.streamer
        except (AttributeError, AssertionError):
            pass  # No task handle proceed
        else:
            self.streamer.stop()
            self.streamer.close()

    def __del__(self):
        if HAS(self.streamer):
            delattr(self, "streamer")


@Provider()
class OnDemandComponent(nidaqmx.Task):

    #: str: name of component
    name = "On-Demand Component"

    # noinspection PyMissingConstructor
    def __init__(self, config: Config):
        self.device_id = config.device_id
        nidaqmx.Task.__init__(self, self.name)

        ################################################################################################################
        # PRE-ALLOCATE
        ################################################################################################################
        #: int: buffer size for each channel
        self.buffer_size = None
        #: np.dtype: dtype for collected data
        self.data_type = None
        #: np.ndarray: buffer object of size channels x buffer size
        self.data = None

        ################################################################################################################
        # INITIALIZATION FROM CONFIG
        ################################################################################################################
        #: int: sampling rate for all DAQ streams (Hz)
        self.sampling_rate = config.sampling_frequency
        #: float: timeout for DAQ interaction (s)
        self.timeout = config.timeout
        #: frozendict: channel index hashmap
        self.channel_index = frozendict({})
        #: frozendict: channel voltage hashmap
        self.channel_limits = frozendict({})
        #: dict: dictionary of individual channel strings
        self.channel_ids = {}
        #: Tuple[str, ...]: key ids of channels
        self.cids = None
        #: int: number of channels
        self.num_channels = 0
        #: str: flattened channel string for DAQ
        self.channel_str = ""
        #: frozendict: channel - buffer row mapping
        self.buffer_index = frozendict({})
        # ingest channel information
        self.retrieve_channel_info(config)

    def __str__(self):
        return self.name

    @abstractmethod
    def command(self, values: Union[float, np.ndarray]) -> "OnDemandComponent":
        ...

    @abstractmethod
    def generate_ni_buffer(self, config):
        ...

    @abstractmethod
    def retrieve_channel_info(self, config: Config) -> "OnDemandComponent":
        """
        Retrieves dictionary of channel index hashmap, channel limits, channel string ids, number of channels,
        and flattened channel string
        """
        ...

    @abstractmethod
    def send(self):
        ...

    @abstractmethod
    def start_component(self) -> OnDemandComponent:
        ...

    def destroy(self):
        try:
            # Make sure it's actually a defined task and not the base component or empty streamer
            assert self._handle
        except (AttributeError, AssertionError):
            pass  # No task handle proceed
        else:
            self.stop()
            self.close()


"""
Analog Input
"""


class AnalogMultiInputStream(StreamingComponent):
    """
    Object for streaming analog inputs
    """
    #: str: name of analog input object
    name = "Analog Input Stream"

    def __init__(self, config: Config):
        self.timestamps = None
        self.voltage_range = config.analog_input_range

        super().__init__(config)
        self.generate_ni_buffer(config)
        self.generate_ring_buffer()
        self.generate_streamer()
        self.generate_processor()

        # make sure this is multi-stream
        if self.num_channels <= 1:
            raise ValueError(f"AnalogMultiInputStream requires multiple channels: {self.num_channels=}")

        # set callbacks
        self.digital_input_request = CallbackRequest(requesting_instance=self, key="digital_input_request")

    def retrieve_channel_info(self, config: Config) -> AnalogMultiInputStream:
        """
        Retrieves dictionary of channel index hashmap, channel string ids, number of channels,
        and flattened channel string
        """
        self.channel_index = config.analog_input_index
        self.channel_ids = config.analog_input_channels
        self.num_channels = config.num_analog_input_channels
        self.channel_str = config.flattened_analog_inputs
        self.buffer_index = frozendict(
            {pair[1]: pair[0] for pair in enumerate(config.analog_input_index.keys())})

    def generate_streamer(self) -> AnalogMultiInputStream:
        self.streamer = AnalogAcquisition(self.name, self.channel_str, self.voltage_range,
                                          self.sampling_frequency)

    def generate_processor(self) -> AnalogMultiInputStream:
        self.stream_processor = AnalogMultiStreamReader(self.streamer)

    def generate_ni_buffer(self, config: "Config"):
        self.buffer_size = config.analog_buffer_size
        self.data_type = np.float64
        self.data = np.full((self.num_channels, self.buffer_size), 0, dtype=self.data_type)

    def start_component(self) -> AnalogMultiInputStream:
        """
        Sets callbacks for every N samples and then starts data acquisition

        :rtype: Self
        """
        self.streamer.register_every_n_samples_acquired_into_buffer_event(
            self.buffer_size, callback_method=self._stream_callback)
        self.streamer.start()

    def _stream_callback_read(self) -> int:
        """
        In place update of a data stream

        :param task_handle: task handle for nidaqmx task
        :type task_handle: str
        :param every_n_samples_event_type: specifies the event is an every_n_samples_event
        :type every_n_samples_event_type: Any
        :param number_of_samples: samples to collect before callback
        :type number_of_samples: int
        :param callback_data: reference to the data buffer
        :type callback_data: np.ndarray
        :return: 0
        :rtype: int
        """
        self.stream_processor.read_many_sample(data=self.data, number_of_samples_per_channel=self.buffer_size,
                                               timeout=self.timeout)

    def _stream_callback_calls(self, data):

        if self.digital_input_request:
            self.digital_input_request()


# noinspection PyUnresolvedReferences
class AnalogAcquisition(nidaqmx.Task):
    """
    Analog Input Source (Streamer)
    """

    def __init__(self, name: str, flattened_channel_str: str, voltage_range: Tuple[float, float], sampling_rate: int):
        # call superclass
        nidaqmx.Task.__init__(self, name)
        # generate channels
        self.ai_channels.add_ai_voltage_chan(flattened_channel_str, name,
                                             terminal_config=nidaqmx.constants.TerminalConfiguration.RSE,
                                             min_val=voltage_range[0], max_val=voltage_range[1])
        # establish timing
        self.timing.cfg_samp_clk_timing(sampling_rate, sample_mode=nidaqmx.constants.AcquisitionType.CONTINUOUS)


class AnalogMultiStreamReader(nidaqmx.stream_readers.AnalogMultiChannelReader):
    """
    Reader of a multi-channel analog data stream

    """
    def __init__(self, analog: AnalogAcquisition):
        super().__init__(analog.in_stream)


class AnalogSingleStreamReader(nidaqmx.stream_readers.AnalogSingleChannelReader):
    """
    Reader of a single-channel analog data stream

    """
    def __init__(self, analog: AnalogAcquisition):
        super().__init__(analog.in_stream)


"""
Analog Output
"""


class AnalogWriter(OnDemandComponent):
    """
    Instance Factory for writing analog

    """
    #: str: name of component
    name = "Analog Output"

    def __init__(self, config: Config):

        # I AM TASK
        # nidaqmx.Task.__init__(self, self.name)

        super().__init__(config)
        self.generate_ni_buffer(config)

        # get general voltage range
        self.voltage_range = (min(min(config.analog_output_range.values())),
                              max(max(config.analog_output_range.values())))

        #: nidaqmx._task_modules.ao_channel_collection.AOChannelCollection: analog output channel object
        self.ao_channels.add_ao_voltage_chan(self.channel_str, self.name, *self.voltage_range)

        # set output command to voltage min
        for cid in self.cids:
            self.data[config.analog_output_index.get(cid), :] = config.analog_output_range.get(cid)[0]

    def command(self, values: Union[float, np.ndarray]) -> "AnalogWriter":
        self.data[..., :] = values

    def generate_ni_buffer(self, config):
        self.buffer_size = config.analog_buffer_size
        self.data_type = np.float64
        self.data = np.full((self.num_channels, self.buffer_size), 0.0, dtype=self.data_type)

    def retrieve_channel_info(self, config: Config) -> "AnalogWriter":
        """
        Retrieves dictionary of channel index hashmap, channel string ids, number of channels,
        and flattened channel string
        """
        self.channel_index = config.analog_output_index
        self.channel_ids = config.analog_output_channels
        self.num_channels = config.num_analog_output_channels
        self.channel_str = config.flattened_analog_outputs
        self.cids = config.analog_output_cids
        self.channel_limits = config.analog_output_range
        self.buffer_index = config.analog_output_index

    def send(self):
        self.write(data=self.data, auto_start=True, timeout=self.timeout)

    def start_component(self) -> "AnalogWriter":
        self.start()


"""
Digital Input
"""


class DigitalMultiInputStream(StreamingComponent):
    """
    Object for streaming digital inputs
    """
    #: str: name of digital input object
    name = "Digital Input Stream"

    def __init__(self, config: Config):
        super().__init__(config)
        self.generate_ni_buffer(config)
        self.generate_ring_buffer()
        self.generate_streamer()
        self.generate_processor()

        # make sure this is multi-stream
        if self.num_channels <= 1:
            raise ValueError(f"DigitalMultiInputStream requires multiple channels: {self.num_channels=}")
        return

    @CallbackRegistry.register()
    def digital_input_request(self):
        self.stream_request()

    def retrieve_channel_info(self, config: Config) -> DigitalMultiInputStream:
        """
        Retrieves dictionary of channel index hashmap, channel string ids, number of channels,
        and flattened channel string
        """
        self.channel_index = config.digital_input_index
        self.channel_ids = config.digital_input_channels
        self.num_channels = config.num_digital_input_channels
        self.channel_str = config.flattened_digital_inputs
        self.buffer_index = frozendict(
            {pair[1]: pair[0] for pair in enumerate(config.digital_input_index.keys())})

    def start_component(self) -> DigitalMultiInputStream:
        """
        starts data acquisition

        """
        self.streamer.start()

    def generate_streamer(self) -> DigitalMultiInputStream:
        self.streamer = DigitalAcquisition(self.name, self.channel_str, self.sampling_frequency, self.buffer_size,
                                           self.timeout)

    def generate_processor(self) -> DigitalMultiInputStream:
        self.stream_processor = DigitalMultiStreamReader(self.streamer)

    def generate_ni_buffer(self, config: "Config"):
        self.buffer_size = config.digital_buffer_size
        self.data_type = np.uint8
        self.data = np.full((self.num_channels, self.buffer_size), 0, dtype=self.data_type)

    def _stream_callback_read(self):
        self.stream_processor.read_many_sample_port_byte(data=self.data,
                                                         number_of_samples_per_channel=self.buffer_size,
                                                         timeout=self.timeout)

    def _stream_callback_calls(self, data):
        # callbacks for updating data
        #for callback, index in self.update_callbacks:
        #    callback([self.buffer_sample_index, data[index, :]])
        pass


class DigitalAcquisition(nidaqmx.Task):
    """
    Digital Input Source

    """

    def __init__(self, name: str, flattened_channel_str: str, sampling_rate: str, buffer_size: int, timeout: int):
        self.sampling_rate = sampling_rate
        self.buffer_size = buffer_size
        self.timeout = timeout
        nidaqmx.Task.__init__(self, name)
        self.di_channels.add_di_chan(flattened_channel_str, name,
                                     nidaqmx.constants.LineGrouping.CHAN_PER_LINE)


class DigitalMultiStreamReader(nidaqmx.stream_readers.DigitalMultiChannelReader):
    """
    Reader of a multi-channel digital data stream

    """
    def __init__(self, digital: DigitalAcquisition):
        super().__init__(digital.in_stream)


class DigitalSingleStreamReader(nidaqmx.stream_readers.DigitalSingleChannelReader):
    """
    Reader of a single-channel digital data stream

    """
    def __init__(self, digital: DigitalAcquisition):
        super().__init__(digital.in_stream)


"""
Digital Output
"""


class DigitalMultiOutputStream(StreamingComponent):
    #: str: name of digital output object
    name = "Digital MultiOutput Stream"

    def __init__(self, config: Config):
        super().__init__(config)
        self.generate_ni_buffer(config)
        self.generate_streamer()
        self.generate_processor()

        # make sure this is multi-stream
        if self.num_channels <= 1:
            raise ValueError(f"DigitalMultiOutputStream requires multiple channels: {self.num_channels=}")
        return

    def retrieve_channel_info(self, config: Config) -> DigitalMultiOutputStream:
        """
        Retrieves dictionary of channel index hashmap, channel string ids, number of channels,
        and flattened channel string
        """
        self.channel_index = config.digital_output_index
        self.channel_ids = config.digital_output_channels
        self.num_channels = config.num_digital_output_channels
        self.channel_str = config.flattened_digital_outputs
        self.buffer_index = frozendict(
            {pair[1]: pair[0] for pair in enumerate(config.digital_output_index.keys())})

    def generate_streamer(self) -> DigitalMultiOutputStream:
        self.streamer = DigitalWriter(self.name,
                                      self.channel_str,
                                      self.sampling_frequency,
                                      self.buffer_size,
                                      self.timeout)

    def generate_processor(self) -> DigitalMultiOutputStream:
        self.stream_processor = DigitalMultiStreamWriter(self.streamer)

    def generate_ni_buffer(self, config: "Config"):
        self.buffer_size = 1
        self.data_type = np.bool_
        self.data = np.full((self.num_channels, self.buffer_size), False, dtype=self.data_type)

    def start_component(self) -> DigitalMultiOutputStream:
        self.streamer.start()
        self.initialize()

    def stop(self):
        self.streamer.stop()

    def initialize(self) -> DigitalMultiOutputStream:
        # format_data(self.data, self.num_channels)
        #self.data[0] = 1
        self.streamer.write(self.data, auto_start=True, timeout=self.timeout)
        # self.stream_processor.write_many_sample_port_byte(self.data, self.timeout)

    @CallbackRegistry.register(alias="digital_output_write")
    def stream_request(self) -> int:
        self.streamer.write(self.data, auto_start=True, timeout=self.timeout)
        return 0

    def _stream_callback_read(self):
        pass

    def _stream_callback_calls(self, data):
        pass


class DigitalWriter(nidaqmx.Task):
    """
    Digital Output Source

    """
    def __init__(self, name: str, flattened_channel_str: str, sampling_rate: str, buffer_size: int, timeout: int):
        self.sampling_rate = sampling_rate
        self.buffer_size = buffer_size
        self.timeout = timeout
        nidaqmx.Task.__init__(self, name)
        self.do_channels.add_do_chan(flattened_channel_str, name,
                                     nidaqmx.constants.LineGrouping.CHAN_PER_LINE)


class DigitalMultiStreamWriter(nidaqmx.stream_writers.DigitalMultiChannelWriter):
    """
    Writer of a multi-channel digital data stream

    """
    def __init__(self, digital: DigitalWriter):
        super().__init__(digital.out_stream)


class DigitalSingleStreamWriter(nidaqmx.stream_writers.DigitalSingleChannelWriter):
    """
    Writer of a single-channel digital data stream

    """
    def __init__(self, digital: DigitalWriter):
        super().__init__(digital.out_stream)


"""
Complete DAQ Object
"""


@Provider()
class DAQ:
    """
    Container component for national instruments data acquisition. Presently, analog_input loops on the main thread but
    only once for every N samples. The actual collection is abstracted through C++ and is hardware-timed. Depending on
    the model of National Instruments DAQ, the digital input might be software or hardware-timed. Hence, the
    analog_input callback calls digital_input to collect a smaller set of samples. Again, the acquisition is abstracted
    through C++. However, the timing is not guaranteed and may contain small amounts of jitter. Therefore, it's best to
    consider the effective acquisition rate for the digital input to be the analog_input's buffer rate. The output
    components analog_output and digital_output are considered "on-demand" components. Their values change only when
    directed and otherwise hold the latest commanded values. Interfaces can be created to facilitate manipulating these
    values in a more meaningful syntax: for example, the linear actuator interface can move the motor in terms of
    physical units instead of voltage. The adapters are dynamically constructed instance attributes and long-term I
    should think about lifting those up to the main level of the controller container. In this way, it could be
    possible to create adapters for input that simply reference the respective channel and might facilitate collect
    by the GUI--as well as hide the actual hardware implementations. Something to think about...

    :param config: The neurobeam configuration
    :param idx: The index of the DAQ in the configuration's component
    """

    def __init__(self, config: Union["Config", "HoistedConfig"]) -> "DAQ":
        ################################################################################################################
        # REFERENCE CONFIG
        ################################################################################################################
        self.config = config

        ################################################################################################################
        # PRE-ALLOCATE
        ################################################################################################################
        self.analog_input = None
        self.analog_output = None
        self.digital_input = None
        self.digital_output = None
        self._running = False

        ################################################################################################################
        # BUILD COMPONENTS
        ################################################################################################################
        self.build()

    def __str__(self) -> str:
        return f"National Instruments DAQ: {self.config.device_id}"

    def __name__(self) -> str:
        return self.config.device_id

    @property
    def running(self) -> bool:
        """
        :Getter: Returns the running status of the DAQ
        :Getter Type: :class:`bool`
        :Setter: This property is read-only
        """
        return self._running

    @property
    def subcomponents(self) \
            -> Tuple[AnalogMultiInputStream, AnalogWriter, DigitalMultiInputStream, DigitalMultiOutputStream]:

        return [getattr(self, subcomponent) for subcomponent in ("analog_input",
                                                                 "analog_output",
                                                                 "digital_input",
                                                                 "digital_output")]

    @property
    def readers(self) -> Tuple[AnalogMultiInputStream, DigitalMultiInputStream]:
        return [getattr(self, subcomponent) for subcomponent in ("analog_input", "digital_input")]

    @property
    def writers(self) -> Tuple[AnalogWriter, DigitalMultiOutputStream]:
        return [getattr(self, subcomponent) for subcomponent in ("analog_output", "digital_output")]

    def build(self) -> "DAQ":
        self.analog_input = AnalogMultiInputStream(self.config) \
            if self.analog_input is None else self.analog_input
        self.analog_output = AnalogWriter(self.config) \
            if self.analog_output is None else self.analog_output
        self.digital_input = DigitalMultiInputStream(self.config) \
            if self.digital_input is None else self.digital_input
        self.digital_output = DigitalMultiOutputStream(self.config) \
            if self.digital_output is None else self.digital_output

    def start(self) -> "DAQ":
        if not self.running:
            for subcomponent in self.subcomponents:
                subcomponent.start_component()
            self._running = True

    def stop(self) -> "DAQ":
        if self.running:
            # Stop readers first since they might potentially call the writers.
            for subcomponent in self.readers:
                subcomponent.streamer.stop()
            for subcomponent in self.writers:
                subcomponent.stop()
            # Make sure to close and save the ring buffers
            for subcomponent in self.readers:
                subcomponent.ring_buffer.close()
            self._running = False

    def destroy(self):
        if self.running:
            self.stop()
        # Destroy components
        for subcomponent in self.subcomponents:
            if subcomponent:
                subcomponent.destroy()

    def __del__(self):
        self.destroy()



from typing import Callable, Any
from functools import update_wrapper
from types import MappingProxyType


def _find_implementation(registry: dict, *args: Any, **kwargs: Any) -> Callable:
    for condition, function in reversed(registry.items()):
        try:
            if condition(*args, **kwargs):
                return function
        except TypeError:
            pass


def _always_true(*args: Any, **kwargs: Any):
    return True


def conditional_dispatch(func: Callable) -> Callable:
    """
    Conditional-dispatch generic function decorator that transforms a function into a generic function whose behavior is
    defined by registered (arbitrary) conditional statements. Closely matches the syntax of functools.singledispatch.

    """
    # implementation registry
    registry = {}

    def dispatch(*args: Any, **kwargs: Any) -> Callable:
        """
        Runs the dispatch algorithm to return the best available implementation
        for the given conditionals registered on the function.

        """
        return _find_implementation(registry, *args, **kwargs)

    def register(conditional: Callable, function: Callable = None) -> Callable:
        if function is None:
            return lambda f: register(conditional, f)
        else:
            registry[conditional] = function
        return function

    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if not args:
            raise TypeError(f'{funcname} requires at least '
                            '1 positional argument')
        return dispatch(*args, **kwargs)(*args, **kwargs)

    funcname = getattr(func, '__name__', 'conditional_dispatch function')
    registry[_always_true] = func
    wrapper.register = register
    wrapper.dispatch = dispatch
    wrapper.registry = MappingProxyType(registry)
    update_wrapper(wrapper, func)
    return wrapper



import warnings
from weakref import ref, proxy, WeakValueDictionary
from abc import ABC, abstractmethod
from typing import Callable, Mapping, Tuple, Any, Optional, List, Union
from types import MethodType
from .exceptions import (CallbackError, AmbiguousCallbackInjectionError, SingularCallbackCollectionError,
                         RequiredCallbackInjectionError, SingularCallbackCollectionWarning, NoOptionalCallbackWarning,
                         NoOptionalCallbackCollectionWarning, RequiredCallbackCollectionError, DispatchError,
                         CallbackKeyWarning, MissingIdentifierError)
from .extensions import CallbackFunction, Component
from .tools import EMPTY, conditional_dispatch, HAS, PatternMatching
from operator import eq
import inspect
from functools import partial, singledispatchmethod
from .registries import CallbackRegistry, Request, ProviderRegistry


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CALLBACK UTILITIES
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


def bind_callback(callback: CallbackFunction, providers: ProviderRegistry,
                  wildcard: Optional[Union[Tuple[str, Any], Tuple[Tuple[str, Any], ...]]] = None) -> partial:
    """
    Bind a callback function to its instance

    :param callback: The callback function to bind
    :param providers: The providers to bind the callback to
    :param wildcard:
    :returns: The bound callback function
    """
    bound_provider = callback.__qualname__.split(".")[0]
    # TODO: This assumes there are not multiple providers with the same name. This is not necessarily true.
    provider = providers.get(bound_provider, wildcard)
    return partial(callback, proxy(provider))


def fake_callback(alias: str) -> CallbackFunction:
    return rename_callback(alias)(lambda *args, **kwargs: None)


def rename_callback(newname: str) -> Callable:
    """
    Decorator that renames a callback function

    :param newname: New callback function name
    :returns: Callback function with new name
    """

    def decorator(f: Callable) -> Callable:
        f.__name__ = newname
        return f

    return decorator


def wrap_callback_collection(collection: Tuple[Callable, ...], name: str = "callback_collection") -> Callable:
    """
    Wrap a collection of callback functions into one callback function called in the order of the input tuple

    :param collection: Tuple of callback functions
    :param name: Name of the wrapped callback function
    :returns: A new callback function that calls each callback in the collection
    """
    @rename_callback(name)
    def wrapped_callback(*args, **kwargs) -> None:  # noqa: ANN001, ANN201, ANN206
        for callback in collection:
            callback(*args, **kwargs)
    return wrapped_callback


class BoundFunction(CallbackFunction):
    """
    A class that represents a bound method
    """

    @staticmethod
    def _check_parameters(callback: Callable) -> bool:
        """
        Check if the callback has the correct parameters

        :param function: The callback function to check
        :returns: Whether the callback is a bound callback
        """
        parameters = inspect.signature(callback).parameters
        if "self" in parameters or "cls" in parameters:
            return True
        else:
            return False


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// REQUEST CONDITIONALS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


# noinspection PyPep8Naming
def IS_CALLBACK(source: CallbackFunction) -> bool:
    """
    Check if the source is a callback function or whether the function within a partial is a callback function

    :param source: The source to check
    :returns: Whether the source is a callback function
    """
    if isinstance(source, CallbackFunction):
        return True
    elif isinstance(source, partial):
        return IS_CALLBACK(source.func)
    else:
        return False


# noinspection PyPep8Naming
def IS_BOUND(source: CallbackFunction) -> bool:
    """
    Check if the source is a bound method

    :param source: The source to check
    :returns: Whether the source is a bound method
    """
    return isinstance(source, BoundFunction)


# noinspection PyPep8Naming
def IS_REGISTRY(source: CallbackRegistry) -> bool:
    """
    Check if the source is a callback registry

    :param source: The source to check
    :returns: Whether the source is a callback registry
    """
    return isinstance(source, CallbackRegistry)


# noinspection PyPep8Naming
def IS_APPROXIMATE(self: "CallbackRequest") -> bool:
    """
    Check if the callback request is approximate

    :param self: The callback request
    :returns: Whether the callback request is approximate
    """
    return self.approximate


# noinspection PyPep8Naming
def IS_REQUIRED(self: "CallbackRequest") -> bool:
    """
    Check if the callback request is required

    :param self: The callback request
    :returns: Whether the callback request is required
    """
    return self.required


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CALLBACK REQUESTS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class _SingleInjection:
    def __init__(self,
                 requesting_instance: Any,
                 key: str,
                 alias: Optional[str] = None,
                 approximate: bool = False,
                 required: bool = False,
                 arguments: Optional[dict[str, Any]] = None,
                 wildcard: Optional[Union[Tuple[str, Any], Tuple[Tuple[str, Any], ...]]] = None):
        self.requesting_instance = requesting_instance
        self.key = key
        self.alias = alias if HAS(alias) else key
        self.approximate = approximate
        self.required = required
        self.arguments = arguments if HAS(arguments) else {}
        self.wildcard = wildcard

    @conditional_dispatch
    def inject(self,
               source: Union[CallbackFunction, CallbackRegistry, None],
               providers: Optional[dict] = None) -> None:
        raise DispatchError(self.requesting_instance.__class__.__name__)

    @inject.register(lambda self, source, providers=None: IS_CALLBACK(source))
    def _inject_callback(self, source: CallbackFunction, providers: Optional[dict] = None) -> None:
        if IS_BOUND(source):
            source = bind_callback(source, providers, self.wildcard)
        if HAS(self.arguments):
            source = partial(source, **self.arguments)
        setattr(self.requesting_instance, self.alias, source)

    @inject.register(lambda self, source, providers=None: IS_REGISTRY(source))
    def _inject_registry(self, source: CallbackRegistry, providers: Optional[dict] = None) -> None:
        try:
            callback = source.get(self.key, approximate=self.approximate)
        except MissingIdentifierError:
            if self.required:
                raise RequiredCallbackInjectionError(self.requesting_instance.__class__.__name__, self.alias)
            else:
                warnings.warn(NoOptionalCallbackWarning(self.requesting_instance.__class__.__name__, self.alias),
                              stacklevel=3)
                callback = fake_callback(self.alias)
        finally:
            # noinspection PyUnboundLocalVariable
            self.inject(callback, providers)

    @inject.register(lambda self, source, providers=None: EMPTY(source))
    def _inject_empty(self, source: None, providers: Optional[dict] = None) -> None:
        if self.required:
            raise RequiredCallbackInjectionError(self.requesting_instance.__class__.__name__, self.alias)
        else:
            warnings.warn(NoOptionalCallbackWarning(self.requesting_instance.__class__.__name__, self.alias),
                          stacklevel=3)
            callback = fake_callback(self.alias)
            self.inject(callback, providers)


class _CollectionInjection:
    def __init__(self,
                 requesting_instance: Any,
                 key: str,
                 alias: str,
                 approximate: bool = False,
                 required: bool = False,
                 arguments: Optional[dict[str, Any]] = None,
                 wildcard: Optional[Union[Tuple[str, Any], Tuple[Tuple[str, Any], ...]]] = None):
        self.requesting_instance = requesting_instance
        self.key = None
        self.alias = None
        self.approximate = None
        self.required = required
        self.wildcard = wildcard

        # parse keys
        if isinstance(key, tuple):
            self.key = key
            self.approximate = False
            if HAS(arguments) and (len(key) != len(arguments)):
                raise AssertionError("The number of keys and arguments must be equal")
                # TODO: Make a custom exception
            if EMPTY(alias):
                raise AssertionError("An alias must be provided for a collection of callbacks")
                # TODO: Make a custom exception
            self.alias = alias
            self.arguments = arguments
        else:
            self.key = key
            self.alias = alias if HAS(alias) else key
            self.approximate = True
            if HAS(self.arguments):
                warnings.warn("beep", stacklevel=3)
                # TODO: Make a custom warning
            self.arguments = None
        self.required = required

    @conditional_dispatch
    def inject(self,
               source: Union[Tuple[CallbackFunction, ...], CallbackRegistry, None],
               providers: Optional[dict] = None) -> None:
        raise DispatchError(self.requesting_instance.__class__.__name__)

    @inject.register(lambda self, source, providers=None: all(map(IS_CALLBACK, source)))
    def _inject_callback(self, source: CallbackFunction, providers: Optional[dict] = None) -> None:
        callbacks = []
        for idx, callback in enumerate(source):
            if IS_BOUND(source):
                callback = bind_callback(source, providers, self.wildcard)
            if HAS(self.arguments):
                callback = partial(source, **self.arguments[idx])
            callbacks.append(callback)
        wrapped_callback = wrap_callback_collection(tuple(callbacks), self.alias)
        setattr(self.requesting_instance, self.alias, wrapped_callback)

    # noinspection PyUnboundLocalVariable
    @inject.register(lambda self, source, providers=None: IS_REGISTRY(source) and IS_APPROXIMATE(self))
    def _inject_registry_approximate(self, source: CallbackRegistry, providers: Optional[dict] = None) -> None:
        try:
            callbacks = source.get_collection(self.key)
        except MissingIdentifierError:
            callbacks = []
        finally:
            if len(callbacks) <= 1 and self.required:
                raise RequiredCallbackCollectionError(self.requesting_instance.__class__.__name__, self.alias)
            elif len(callbacks) == 0 and not self.required:
                callbacks = [fake_callback(self.alias), fake_callback(self.alias)]
            elif len(callbacks == 1) and not self.required:
                warnings.warn(SingularCallbackCollectionWarning(self.requesting_instance.__class__.__name__,
                                                                self.alias),
                              stacklevel=3)
                callbacks = [callbacks[0], fake_callback(self.alias)]
            self.inject(callbacks, providers)

    @inject.register(lambda self, source, providers=None: IS_REGISTRY(source) and not IS_APPROXIMATE(self))
    def _inject_registry(self, source: CallbackRegistry, providers: Optional[dict] = None) -> None:
        callbacks = []
        for callback in self.key:
            try:
                callbacks = source.get(self.key, approximate=False)
            except MissingIdentifierError:
                if self.required:
                    raise RequiredCallbackCollectionError(self.requesting_instance.__class__.__name__, self.alias)
        if len(callbacks) <= 1 and self.required:
            raise RequiredCallbackCollectionError(self.requesting_instance.__class__.__name__, self.alias)
        elif len(callbacks) == 0 and not self.required:
            callbacks = [fake_callback(self.alias), fake_callback(self.alias)]
        elif len(callbacks == 1) and not self.required:
            warnings.warn(SingularCallbackCollectionWarning(self.requesting_instance.__class__.__name__, self.alias),
                          stacklevel=3)
            callbacks = [callbacks[0], fake_callback(self.alias)]
        self.inject(callbacks, providers)

    @inject.register(lambda self, source, providers=None: EMPTY(source))
    def _inject_empty(self, source: None, providers: Optional[dict] = None) -> None:
        if self.required:
            raise RequiredCallbackInjectionError(self.requesting_instance.__class__.__name__, self.alias)
        else:
            warnings.warn(NoOptionalCallbackWarning(self.requesting_instance.__class__.__name__, self.alias),
                          stacklevel=3)
            callbacks = [fake_callback(self.alias), fake_callback(self.alias)]
            self.inject(callback, providers)


@Request()
class CallbackRequest:
    def __init__(self,
                 requesting_instance: Any,
                 key: Union[str, Tuple[str, ...]],
                 alias: Optional[str] = None,
                 approximate: bool = False,
                 collection: bool = False,
                 required: bool = False,
                 arguments: Optional[Union[dict[str, Any], Tuple[dict[str, Any], ...]]] = None,
                 wildcard: Union[Tuple[str, Any], Tuple[Tuple[str, Any], ...]] = None):
        ##############################################################################################################
        # INSTANTIATE IMPLEMENTATION
        ################################################################################################################
        collection_ = True if collection or isinstance(key, tuple) else False
        if collection_:
            self._implementation = _CollectionInjection(requesting_instance,
                                                        key,
                                                        alias,
                                                        approximate,
                                                        required,
                                                        arguments,
                                                        wildcard)
        else:
            self._implementation = _SingleInjection(requesting_instance,
                                                    key,
                                                    alias,
                                                    approximate,
                                                    required,
                                                    arguments,
                                                    wildcard)

            ############################################################################################################
            # ENSURE ALIAS RESERVED
            ############################################################################################################
            setattr(self._implementation.requesting_instance, self._implementation.alias, None)

    def __str__(self) -> str:
        return f"Request to inject callback {self.alias} into {self.requesting_instance.__class__.__name__}"

    @property
    def alias(self) -> str:
        return self._implementation.alias

    @property
    def approximate(self) -> bool:
        return self._implementation.approximate

    @property
    def collection(self) -> bool:
        return True if isinstance(self._implementation, _CollectionInjection) else False

    @property
    def key(self) -> Union[str, Tuple[str, ...]]:
        return self._implementation.key

    @property
    def requesting_instance(self) -> Any:
        return self._implementation.requesting_instance

    @property
    def required(self) -> bool:
        return self._implementation.required

    def warn_if_alias_unreserved(self) -> None:
        if not hasattr(self._implementation.requesting_instance, self._implementation.alias):
            warnings.warn(CallbackKeyWarning(self._implementation.requesting_instance.__class__.__name__,
                                             self._implementation.alias),
                          stacklevel=2)

    def inject(self,
               source: Union[CallbackFunction, Tuple[CallbackFunction, ...], CallbackRegistry, None],
               providers: Optional[ProviderRegistry] = None) -> None:
        self._implementation.inject(source, providers)

    def __call__(self, *args, **kwargs):
        raise CallbackError(self.requesting_instance.__class__.__name__, self.alias)


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// ADAPTER REQUESTS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@Request()
class AdapterRequest:
    def __init__(self,
                 requesting_instance: Any,
                 key: str,
                 alias: Optional[str] = None,
                 wildcard: Optional[Tuple[str, Any]] = None):
        self.requesting_instance = requesting_instance
        self.key = key
        self.alias = alias if HAS(alias) else key
        self.wildcard = wildcard

    def __str__(self) -> str:
        return f"Request to inject {self.key} instance into {self.requesting_instance.__class__.__name__}"

    def inject(self, providers: ProviderRegistry) -> None:
        provider = providers.get(self.key, self.wildcard)
        setattr(self.requesting_instance, self.alias, provider)
        setattr(self.requesting_instance, "initialized", True)


@Request()
class TriggerRequest:
    def __init__(self,
                 key: str,
                 wildcard: Optional[Tuple[str, Any]] = None):
        self.key = key
        self.wildcard = wildcard

    def __str__(self) -> str:
        return f"Request to inject {self.key} instance into {self.requesting_instance.__class__.__name__}"

    def inject(self, providers: ProviderRegistry) -> None:
        return providers.get(self.key, self.wildcard)



from sys import float_info as _float_info
from typing import Any, Callable, Optional, Union
from random import shuffle

import numpy as np

from ..configs import CallbackRequestConfig
from ..timing import Seconds, Time, timer
from ..requests import fake_callback, CallbackRequest
from ..registries import BehaviorRegistry, Provider
from ..requests import fake_callback, CallbackRequest, TriggerRequest
from ..tools import HAS


# constants used for defaults
MIN = _float_info.min
MAX = _float_info.max


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Standard Base State
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@BehaviorRegistry.register()
@Provider()
class State:
    """
    State within the behavioral state machine.

    :param name: name of state
    :param event_log: event log for state
    :param duration: duration of state before auto-transition
    :param entry_callbacks: callbacks for entry into state
    :param exit_callbacks: callbacks for exit from state
    :param triggers: triggers for state
    """

    def __init__(self,
                 name: str,
                 event_log: Optional[Callable] = None,
                 duration: Optional["Time"] = None,
                 entry_callbacks: Optional[Union[CallbackRequestConfig, CallbackRequest, Callable]] = None,
                 exit_callbacks: Optional[Union[CallbackRequestConfig, CallbackRequest, Callable]] = None,
                 triggers: Optional = None,
                 ):
        ################################################################################################################
        # GENERAL CONFIGURATION
        ################################################################################################################
        #: str: name of state
        self.name = name
        #: "Time": duration of state before an auto-transition
        self.duration = duration
        #: Optional: triggers in the state
        self.triggers = triggers

        if HAS(triggers):
            self.triggers = [TriggerRequest(key=key, wildcard=wildcard) for key, wildcard in triggers]

        ################################################################################################################
        # CALLBACKS
        ################################################################################################################
        #: bool: whether entry/exit callbacks are currently wrapped for event logging
        self.__callbacks_wrapped = False
        if HAS(entry_callbacks) and isinstance(entry_callbacks, CallbackRequestConfig):
            entry_callbacks = CallbackRequest(requesting_instance=self,
                                              key=entry_callbacks.key,
                                              alias="__entry_callback__",
                                              approximate=entry_callbacks.approximate,
                                              required=entry_callbacks.required,
                                              arguments=entry_callbacks.arguments)
        if HAS(exit_callbacks) and isinstance(exit_callbacks, CallbackRequestConfig):
            exit_callbacks = CallbackRequest(requesting_instance=self,
                                             key=exit_callbacks.key,
                                             alias="__exit_callback__",
                                             approximate=exit_callbacks.approximate,
                                             required=exit_callbacks.required,
                                             arguments=exit_callbacks.arguments)

        self.__entry_callback__ = entry_callbacks if HAS(entry_callbacks) \
            else fake_callback("__entry_callback__")  # noqa: U100
        self.__exit_callback__ = exit_callbacks if HAS(exit_callbacks) \
            else fake_callback("__exit_callback__")  # noqa: U100

        ################################################################################################################
        # LOGGING
        ################################################################################################################
        self.__log_event__ = lambda *args: None  # noqa: U100
        self.__event_wrapper__ = lambda func: func  # noqa: U100
        if event_log:
            self.__log_event__ = event_log.log_event
            self.__event_wrapper__ = event_log.wrap

        ################################################################################################################
        # PRE-ALLOCATE
        ################################################################################################################
        #: int: number of entries into state
        self.entries = 0
        #: int: number of exits from state
        self.exits = 0
        #: float: start of state
        self.start_time = MIN
        #: float: auto-transition time
        self.auto_time = MAX
        #: float: progress of state
        self.state_progress = 0.0

        ################################################################################################################
        # SET CHECK FUNCTIONS
        ################################################################################################################
        self.check_progress = self.check_progress if HAS(self.duration) else lambda time: 0.0  # noqa: U100
        self.check_transition = self.check_transition if HAS(self.duration) else lambda time: False  # noqa: U100
        self.check_triggers = self.check_triggers if HAS(self.triggers) else lambda time: None  # noqa: U100

    @property
    def callbacks_wrapped(self) -> bool:
        return self.__callbacks_wrapped

    def wrap_callbacks(self) -> None:
        if not self.__callbacks_wrapped:
            self.__entry_callback__ = self.__event_wrapper__(self.__entry_callback__)
            self.__exit_callback__ = self.__event_wrapper__(self.__exit_callback__)
            self.__callbacks_wrapped = True

    def __str__(self):
        return f"{self.name} state with {self.entries} entries and {self.exits} exits"

    def state_loop(self, time: float) -> bool:
        self.state_progress = self.check_progress(time)
        self.check_triggers(time)
        return self.check_transition(time)

    def check_triggers(self) -> None:
        for trigger in self.triggers:
            trigger.trigger(time)

    def check_progress(self, time: float) -> float:
        return ((time - self.start_time) / self.duration) * 100

    def check_transition(self, time: float) -> bool:
        return time >= self.auto_time

    def initialize(self) -> "State":
        """
        Initialize state
        """
        self.start_time = timer()
        self.entries += 1
        self.__log_event__(self.name, "entry")

        # set duration
        if HAS(self.duration):
            self.auto_time = self.start_time + self.duration
        # TODO POTENTIAL OVERFLOW???

        if isinstance(self.triggers, list):
            for trigger in self.triggers:
                trigger.initialize(self.start_time)

        # run entry callback
        self.__entry_callback__()

    def transition(self) -> "State":
        self.exits += 1
        self.__exit_callback__()
        self.__log_event__(self.name, "exit")

    def __eq__(self, other: Any):
        return self.name == other

    def __ne__(self, other: Any):
        return self.name != other

    def __hash__(self):
        return id(self)

    def __name__(self):
        return f"State: {self.name}"


@BehaviorRegistry.register()
class Setup(State):
    #: str: key for event log (debugging)
    _event_log_key = "event_log"

    def __init__(self, event_log: Optional[Union[str, Callable]] = None, **kwargs):
        super().__init__(name="Setup", event_log=event_log)

    def initialize(self) -> "State":
        if self.entries != 0:
            raise InvalidStateEntryError(self.name)
        super().initialize()


@BehaviorRegistry.register()
class End(State):
    def __init__(self, event_log: Optional[Callable] = None, **kwargs):

        final_callback = CallbackRequest(self,
                                         "reporting_console_task_finished",
                                         alias="__entry_callback__")

        super().__init__(name="End", event_log=event_log, entry_callbacks=final_callback)

    def initialize(self) -> "State":
        if self.entries != 0:
            raise InvalidStateEntryError(self.name)
        super().initialize()


@BehaviorRegistry.register()
class Retract(State):
    def __init__(self, event_log: Optional[Callable] = None, **kwargs):
        entry_callback = CallbackRequest(self,
                                         "linear_actuator_command",
                                         alias="__entry_callback__",
                                         required=True,
                                         arguments={"value": 0.0}
                                         )
        duration = Seconds(5)
        super().__init__(name="Retract", entry_callbacks=entry_callback, duration=duration, event_log=event_log)


@BehaviorRegistry.register()
class Release(State):
    def __init__(self, event_log: Optional[Callable] = None, **kwargs):
        entry_callback = CallbackRequest(self,
                                         "linear_actuator_command",
                                         alias="__entry_callback__",
                                         required=True,
                                         arguments={"value": 100.0})
        duration = Seconds(5)
        super().__init__(name="Release", entry_callbacks=entry_callback, duration=duration, event_log=event_log)


@BehaviorRegistry.register()
@Provider()
class Selector:
    def __init__(self,
                 name: str,
                 includes: dict[str, int],
                 final: str,
                 event_log: Optional[Callable] = None,
                 **kwargs
                 ):
        ######################################################################################################
        # INITIALIZATION
        ######################################################################################################
        #: str: name of selector
        self.name = name
        self.index = []
        assert len(set(includes.values())) == 1, "All includes must have the same number of trials"
        num_blocks = list(includes.values())[0]
        unique_states = list(range(len(includes.keys())))

        for _ in range(num_blocks):
            block = unique_states.copy()
            shuffle(block)
            self.index.extend(block)

        #for idx, kv in enumerate(includes.items()):
        #    key, value = kv
        #    for _ in range(value):
        #        self.index.append(idx)
        #shuffle(self.index)
        self.index.append(len(includes.keys()))
        ######################################################################################################
        # LOGGING
        ######################################################################################################
        self.__log_event__ = lambda *args: None  # noqa: U100

        ######################################################################################################
        # PRE-ALLOCATE
        ######################################################################################################
        #: int: number of entries into state
        self.entries = 0
        #: int: number of exits from state
        self.exits = 0
        self.state_progress = 0

    def __str__(self):
        return f"{self.name}({self.index})"

    def state_loop(self, time: float) -> bool:
        return True

    def initialize(self):
        self.entries += 1
        self.__log_event__(self.name, "entry")

    def transition(self) -> "State":
        self.exits += 1
        self.__log_event__(self.name, "exit")
        try:
            return self.index[self.exits - 1]
        except IndexError:
            print(f"{self=}")
            print(f"{self.index=}")
            print(f"{self.entries=}")
            print(f"{self.exits=}")

    def __eq__(self, other: Any):
        return self.name == other

    def __ne__(self, other: Any):
        return self.name != other

    def __name__(self):
        return f"State: {self.name}"

    @property
    def callbacks_wrapped(self) -> bool:
        return True

    def wrap_callbacks(self) -> None:
        pass


@BehaviorRegistry.register()
class VariableState(State):
    def __init__(self,
                 name: str,
                 duration: "Time",
                 statistics: dict,
                 event_log: Optional[Callable] = None,
                 entry_callbacks:  Optional[Union[CallbackRequestConfig, CallbackRequest, Callable]] = None,
                 exit_callbacks:  Optional[Union[CallbackRequestConfig, CallbackRequest, Callable]] = None,
                 triggers: Optional[Union[CallbackRequestConfig, CallbackRequest, Callable]] = None,
                 **kwargs
                 ):

        if "std" not in statistics:
            warnings.warn("NO STD")

        self.duration_std = statistics.get("std")
        self.duration_mean = duration
        self.duration_lower_limit = statistics.get("lower_limit", 0)
        self.duration_upper_limit = statistics.get("upper_limit", MAX)

        super().__init__(name, event_log, duration, entry_callbacks, exit_callbacks, stimulus)

    def initialize(self) -> "VariableState":
        self.duration = np.random.normal(self.duration_mean, self.duration_std)
        self.duration = max(self.duration, self.duration_lower_limit)
        self.duration = min(self.duration, self.duration_upper_limit)
        super().initialize()



import warnings
from typing import Any, Tuple, Union, Optional, Callable, Mapping, List
from functools import partial, singledispatchmethod
from frozendict import frozendict
import numpy as np
from pathlib import Path

from ..configs import HoistedConfig, StateConfig, TransitionConfig
from ..extensions import BehaviorState
from ..data import EventLog
from ..exceptions import (InvalidStateEntryError, DuplicateRegistrationError, MissingIdentifierError)
from ..factories import BehaviorFactory
from ..registries import ComponentRegistry, Provider
from ..requests import CallbackRequest, wrap_callback_collection, fake_callback
from ..timing import Seconds, timer, Time
from ..tools import HAS, FORMAT_TERMINAL, prune_keys, parse_empties


from sys import float_info as _float_info


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// BEHAVIOR
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@ComponentRegistry.register()
@Provider()
class Behavior:
    """
    Behavioral component.
    """
    #: str: key for event log (debugging)
    _event_log_key = "event_log"

    def __init__(self,
                 config: Optional["HoistedConfig"] = None,
                 name: str = "Behavioral Task",
                 states: Tuple["BehaviorState", ...] = (StateConfig(), ),
                 transitions: Tuple["TransitionConfig", ...] = (TransitionConfig(), ),
                 reporting_interval: float = 1.0,
                 output_folder: Path = Path.cwd()):
        ################################################################################################################
        # CONFIGURE
        ################################################################################################################
        #: "Config": configuration
        self.config = config
        output_folder = self.config.save_location if HAS(config) else output_folder
        #: "EventLog": Logs events of the behavior
        self.event_log = EventLog(key=self._event_log_key, path=output_folder, ring_size=5)
        # TODO: Remove hardcode for ring size of event log
        #: str: name of behavior
        self.name = self.config.name if HAS(config) else name
        #: Tuple["BehaviorState"]: states
        self.states = self.config.states if HAS(config) else states
        #: Tuple["TransitionConfig"]: transitions
        self.transitions = self.config.transitions if HAS(config) else transitions
        #: float: reporting interval
        self.reporting_interval = self.config.reporting_interval if HAS(config) else reporting_interval

        ################################################################################################################
        # PRE-ALLOCATE
        ################################################################################################################
        #: "_BehavioralStateMachine": state machine implementing behavior
        self.machine = None

        ################################################################################################################
        # BUILD
        ################################################################################################################
        self.build()

    def __str__(self):
        return self.name
        # TODO: Implement more informative __str__ for behavior

    def build(self) -> "Behavior":
        self.machine = BehaviorFactory.construct(states=self.states,
                                                 transitions=self.transitions,
                                                 event_log=self.event_log,
                                                 reporting_interval=self.reporting_interval)
        self.event_log.log_event("Behavior", "built")

    def start(self) -> "Behavior":
        if not self.machine.running:
            self.event_log.log_event("Behavior", "start")
            self.machine.start()

    def stop(self) -> "Behavior":
        if self.machine.running:
            self.machine.stop()
            self.event_log.log_event("Behavior", "stop")
        # self.event_log.close()

    def destroy(self) -> None:
        self.event_log.close()

    def __name__(self):
        return self.name



from typing import Tuple, Callable, List, Any
from collections import deque
from pathlib import Path
from weakref import WeakValueDictionary
import warnings
from threading import RLock
from functools import partial

import numpy as np

from .exceptions import (OverloadedEventLogWarning, EmptyEventLogWarning, PathMismatchWarning,
                         RingSizeMismatchWarning, DataReorganizationWarning, DesynchronizedBuffersWarning)
from .timing import timer
from .tools import conditional_dispatch, EMPTY, HAS


"""
A collection of classes that simultaneously write data to file and maintain a ring buffer in memory for access by other
classes.
"""


class EventLog(deque):
    """
    Fast, fixed-size FIFO thread-safe ring buffer for event logs with concurrent writing to file. Implemented as
    doubly-linked list for O(1) append and pop operations. Implemented using a multiton design pattern allowing us to
    attach it to multiple threads and classes without explicit passing, but allows creation or selection of distinct
    instances using an instance cache. Calling add_data will add a timestamped event to the log. To avoid a ton of
    persistent entry-points or references to a single instance, you can use the wrap method to wrap the log across a
    function and remove the local instance of the log. Calling get-data will return the logged events with the
    rightmost element being the most recent.

    :param key: The key for instance registry
    :param path: The path that will have the key appended as a filename with extension .log
        default is the current working directory but not annotated as such
    :param ring_size: The number of events to store in the ring buffer log; default is 1 but not annotated as such
    :raises PathMismatchWarning: Raised if path does not match existing instance
    :raises RingSizeMismatchWarning: Raised if ring_size does not match existing instance
    :raises EmptyEventLogWarning: Raised if no event and action is provided to write data
    :raises OverloadedEventLogWarning: Raised if more than two arguments are provided to write data

    .. note :: This class adheres to the structural requirements of
        :class:`DataBuffer <RF_PROBE.extensions.DataBuffer>`

    .. seealso:: https://en.wikipedia.org/wiki/Multiton_pattern
    """

    #: WeakValueDictionary: registry of instances
    __instance_registry = WeakValueDictionary()

    #: RLock: recursive mutex for thread safety
    __recursive_mutex = RLock()

    #: slots for instance attributes (faster attribute access)
    __slots__ = [
        "file",  #: file: file-handle
        "path",  #: Path: path to log file
        "_is_open"  #: bool: whether file is open
    ]

    def __new__(cls, key: str = "event_log", path: Path = None, ring_size: int = None) -> "EventLog":
        """
        Fast, fixed-size FIFO thread-safe ring buffer for event logs with concurrent writing to file. Implemented as
        doubly-linked list for O(1) append and pop operations. Implemented using a multiton design pattern allowing us
        to attach it to multiple threads and classes without explicit passing, but allows creation or selection of
        distinct instances using an instance cache. Calling add_data will add a timestamped event to the log. To avoid
        a ton of persistent entry-points or references to a single instance, you can use the wrap method to wrap the
        log across a function and remove the local instance of the log. Calling get-data will return the logged events
        with the rightmost element being the most recent.

        :param key: The key for instance registry
        :param path: The path that will have the key appended as a filename with extension .log
            default is the current working directory but not annotated as such
        :param ring_size: The number of events to store in the ring buffer log; default is 1 but not annotated as such
        :raises PathMismatchWarning: Raised if path does not match existing instance
        :raises RingSizeMismatchWarning: Raised if ring_size does not match existing instance
        :raises EmptyEventLogWarning: Raised if no event and action is provided to write data
        :raises OverloadedEventLogWarning: Raised if more than two arguments are provided to write data

        .. note :: This class adheres to the structural requirements of
            :class:`DataBuffer <RF_PROBE.extensions.DataBuffer>`

        .. seealso:: https://en.wikipedia.org/wiki/Multiton_pattern
        """
        if EMPTY(cls.__instance_registry):
            return super().__new__(cls)
        elif key in cls.__instance_registry:
            instance = cls.__instance_registry.get(key)
            if HAS(path) and path != instance.path:
                warnings.warn(PathMismatchWarning(key, path, instance.path), stacklevel=2)
            if HAS(ring_size) and ring_size != len(instance):
                warnings.warn(RingSizeMismatchWarning(key, ring_size, len(instance)), stacklevel=2)
            return instance
        else:
            return super().__new__(cls)

    def __init__(self, key: str = "event_log", path: Path = None, ring_size: int = None):
        """
        Fast, fixed-size FIFO thread-safe ring buffer for event logs with concurrent writing to file. Implemented as
        doubly-linked list for O(1) append and pop operations. Implemented using a multiton design pattern allowing us
        to attach it to multiple threads and classes without explicit passing, but allows creation or selection of
        distinct instances using an instance cache. Calling add_data will add a timestamped event to the log. To avoid
        a ton of persistent entry-points or references to a single instance, you can use the wrap method to wrap the
        log across a function and remove the local instance of the log. Calling get-data will return the logged events
        with the rightmost element being the most recent.

        :param key: The key for instance registry
        :param path: The path that will have the key appended as a filename with extension .log
            default is the current working directory but not annotated as such
        :param ring_size: The number of events to store in the ring buffer log; default is 1 but not annotated as such
        :raises PathMismatchWarning: Raised if path does not match existing instance
        :raises RingSizeMismatchWarning: Raised if ring_size does not match existing instance
        :raises EmptyEventLogWarning: Raised if no event and action is provided to write data
        :raises OverloadedEventLogWarning: Raised if more than two arguments are provided to write data

        .. note :: This class adheres to the structural requirements of
            :class:`DataBuffer <RF_PROBE.extensions.DataBuffer>`

        .. seealso:: https://en.wikipedia.org/wiki/Multiton_pattern
        """
        ################################################################################################################
        # ONLY INITIALIZE IF KEY NOT IN REGISTRY
        ################################################################################################################
        if key not in self.__instance_registry:
            if EMPTY(ring_size):
                ring_size = 1  # we provide a default this way so that we can warn the user so that we can
                # more simply identify if instance registry is called with a ring size mismatch
            super().__init__(list(range(ring_size)))
            if EMPTY(path):  # we provide a default this way so that we can warn the user we're reverting to default
                path = Path.cwd()
                warnings.warn(f"No path provided for {key}.log. Writing to {path}", stacklevel=2)
            # initialize instance attributes
            self.file = None
            self._is_open = False
            self.path = path.joinpath(f"{key}.log")
            self.open()
            self._initialize_data()
            self.__instance_registry[key] = self  # Add to registry only after successful initialization

    def __str__(self):
        return f"{self.__call__()}"

    @property
    def is_open(self) -> bool:
        """
        :Getter: Whether the log file is open.
        :Getter Type: :class:`bool`
        :Setter: This property is read-only.
        """
        return self._is_open

    def add_data(self, event: str, action: str) -> "EventLog":
        """
        Adds an event to the log with a timestamp in the form of "Timestamp\tEvent\tAction" with a tab delimiter. The
        timestamp is in microseconds with one decimal place. The timestamp is not directly related to any wall-clock,
        and will need to be compared relative to other timestamps to be useful. The event is the name of the
        object  (class, function, etc.) being logged, and the action is some description of the event. For example,
        "3504856465367.2\tButton\tPressed". This method is thread-safe.

        :param event: The name of the object or function being logged
        :param action: The action being logged

        .. warning:: The timestamp is not directly related to any wall-clock, and will need to be compared relative to
            other timestamps to be useful. The easiest way of doing this is to call a timestamp at the beginning of the
            experiment and use the timestamp as time zero.

        .. seealso:: :class:`Microseconds <RF_PROBE.timing.Microseconds>`
        """
        with self.__recursive_mutex:  # technically we make writing safe in _write data, but I'd prefer to only have
            # one thread calling add_data at a time. We'll use a reentrant lock
            data = f"{timer()}\t{event}\t{action}\n"  # format the data for writing
            self._write_data(data)
            self.pop()
            self.appendleft(data)

    def clear(self) -> "EventLog":
        """
        Clears the log's ring buffer and file handle
        """
        if self.is_open:
            self.close()
        self.file = None
        super().clear()

    def close(self) -> "EventLog":
        """
        Closes the log file.
        """
        with self.__recursive_mutex:
            if self.is_open:
                self.file.close()
                self._is_open = False

    def get_data(self) -> List[str]:
        """
        Retrieve the logged events the n-most recent logged events, where n is the size of the ring buffer
        (i.e., "ring_size").

        :returns: Logged events in the form of a list of strings where each string is in the form of
            "Timestamp\tEvent\tAction" with a tab delimiter. The rightmost element is the most recent.
        """
        return self  # this call is only here for consistency with other buffers

    def open(self) -> "EventLog":
        """
        Opens the log file.
        """
        if not self.is_open:
            self.file = open(self.path, "w")
            self._is_open = True

    def wrap(self, func: Callable) -> Callable:
        """
        Decorator method that wraps a function or method with the log_event method. This method allows one to log the
        calling of  an event without having to explicitly call the log_event method each time.

        :param func: The function to be decorated
        :returns: The new decorated function or method that can be used in place of the original one.
        """
        def wrapped_function(*args, **kwargs):  # noqa: ANN201
            if isinstance(func, partial):
                name = func.func.__name__
            else:
                name = func.__name__
            self.log_event(name, "called")
            return func(*args, **kwargs)
        return wrapped_function

    @conditional_dispatch
    def log_event(self, *args) -> "EventLog":
        """
        Dispatch method for logging events. This method is overloaded to handle different numbers of arguments. One
        could use a single method with "NULL" for default values of event and action, but the intention here is to
        provide a warning if the caller is not providing any information to the log rather than intentionally
        providing NULL.

        :param args: The first position argument is a string representing the event, and the second position argument
            is a string representing the action. If only one argument is provided, the action is set to "NULL".
        :raises EmptyEventLogWarning: Raised if no event or action is provided. In this case, the event and action are
            set to "NULL".
        :raises OverloadedEventLogWarning: Raised if more than two arguments have been provided. In this case, only the
            first two arguments will be used.
        """
        ...

    @log_event.register(lambda *args: len(args) == 1)
    def _(self, *args) -> "EventLog":
        """
        Method used when only the reference to the EventLog instance is provided. Sets both event and action to "NULL".

        :raises EmptyEventLogWarning:
        """
        self.add_data("NULL", "NULL")
        warnings.warn(EmptyEventLogWarning(), stacklevel=2)

    @log_event.register(lambda *args: len(args) == 2)
    def _(self, *args) -> "EventLog":
        """
        Method used when only the event is provided. Sets the action to "NULL".
        """
        self.add_data(args[0], "NULL")

    @log_event.register(lambda *args: len(args) == 3)
    def _(self, *args) -> "EventLog":
        """
        Method used when both the event and action are provided.
        """
        self.add_data(*args)

    @log_event.register(lambda *args: len(args) >= 4)
    def _(self, *args) -> "EventLog":
        """
        Method used when more than two arguments are provided. Only the first two arguments are used.

        :raises OverloadedEventLogWarning:
        """
        self.add_data(*args[:2])
        warnings.warn(OverloadedEventLogWarning(args), stacklevel=2)

    def _initialize_data(self) -> "EventLog":
        """
        Initializes the log's ring buffer with "Timestamp\tNULL\tNULL\n" for each element in the ring buffer.
        """
        for _ in range(len(self)):
            self.pop()
            self.appendleft(f"{timer()}\tNULL\tNULL\n")

    def _write_data(self, data: str) -> "EventLog":
        """
        Logs an event to the log file. This method is thread-safe.

        :param data: A string in the format of "Timestamp\tEvent\tAction" with a tab delimiter.
        """
        with self.__recursive_mutex:
            self.file.write(data)
            self.file.flush()


class RingBuffer(deque):
    """
    Fast, fixed-size FIFO thread-safe ring buffer for numerical data with concurrent writing to file. Implemented as
    doubly-linked list for O(1) append and pop operations.

    :param element_depth: The number of rows in each block of data
    :param element_length: The number of columns in each block of data
    :param ring_size: The number of blocks of data to store in the ring buffer
    :param path: The path to the file where the data will be written
    :param dtype: The data type of the data to be written
    :param do_reorganize: Whether to reorganize the data into a single numpy array after writing; default is True
    :raises DataReorganizationWarning: Raised if the number of elements in the raw data file does not match the number
     of elements required to reorganize the data. In this case, the raw data file will not be deleted.

    .. note :: This class adheres to the structural requirements of
        :class:`DataBuffer <RF_PROBE.extensions.DataBuffer>`

    """
    __slots__ = [
        "pointer",
        # "block_shape", # Turns out that using slots prevents the use of cached_properties
        # "block_size", # I actually think it might be premature optimization to incorporate
        # "data_shape", # a property that calls a cached function, so commenting out for now
        # "data_size", # and leaving the properties as standard
        "element_depth",
        "element_length",
        "raw_path",
        "org_path",
        "data_type",
        "file",
        "do_reorganize",
        "_is_open"
    ]

    def __init__(self,
                 element_depth: int,
                 element_length: int,
                 ring_size: int,
                 path: Path,
                 dtype: np.dtype = np.float64,
                 do_reorganize: bool = True):
        """
        Fast, fixed-size FIFO thread-safe ring buffer for numerical data with concurrent writing to file. Implemented
        as doubly-linked list for O(1) append and pop operations.

        :param element_depth: The number of rows in each block of data
        :param element_length: The number of columns in each block of data
        :param ring_size: The number of blocks of data to store in the ring buffer
        :param path: The path to the file where the data will be written
        :param dtype: The data type of the data to be written
        :param do_reorganize: Whether to reorganize the data into a single numpy array after writing; default is True
        :raises DataReorganizationWarning: Raised if the number of elements in the raw data file does not match the
            number of elements required to reorganize the data. In this case, the raw data file will not be deleted

        .. note :: This class adheres to the structural requirements of
            :class:`DataBuffer <RF_PROBE.extensions.DataBuffer>`
        """
        super().__init__(list(range(ring_size)))
        self._is_open = False
        self.file = None
        self.do_reorganize = do_reorganize
        #: Path: path to for raw data
        if self.do_reorganize:
            self.raw_path = Path("".join([str(path), "_raw"]))
        else:
            self.raw_path = path
        #: Path: path for organized data
        self.org_path = path
        #: numpy.dtype: data type
        self.data_type = dtype
        #: int: index containing location for the most block of data to add to the ring  in the written data
        self.pointer = 0
        #: int: rows in data
        self.element_depth = element_depth
        #: int: columns in data
        self.element_length = element_length
        #: file-handle
        self.open()
        self._initialize_data()

    def __str__(self):
        return f"{self.__call__()}"

    @property
    def block_shape(self) -> Tuple[int, int]:
        """
        :Getter: The shape of each block of data in the ring buffer.
        :Getter Type: :class:`Tuple <tuple>`\\[:class:`int`, :class:`int`\\]
        :Setter: This property is read-only.
        """
        return self.element_depth, self.element_length

    @property
    def block_size(self) -> int:
        """
        :Getter: The size of each block of data in the ring buffer.
        :Getter Type: :class:`int`
        :Setter: This property is read-only.
        """
        return self.element_depth * self.element_length

    @property
    def data_shape(self) -> Tuple[int, int]:
        """
        :Getter: The shape of the data in the organized file.
        :Getter Type: :class:`Tuple <tuple>`\\[:class:`int`, :class:`int`\\]
        :Setter: This property is read-only.
        """
        return self.element_depth, self.element_length * self.pointer

    @property
    def data_size(self) -> int:
        """
        :Getter: The size of the data in the organized file.
        :Getter Type: :class:`int`
        :Setter: This property is read-only.
        """
        return self.element_depth * self.element_length * self.pointer

    @property
    def is_open(self) -> bool:
        """
        :Getter: Whether the associated file is open.
        :Getter Type: :class:`bool`
        :Setter: This property is read-only.
        """
        return self._is_open

    def add_data(self, data: np.ndarray) -> "RingBuffer":
        """
        Adds data to both the ring buffer and to the data file.

        :param data: The data to be added.
        """
        self._write_data(data)
        self.pop()
        self.appendleft(data)
        self.pointer += 1

    def clear(self) -> "RingBuffer":
        """
        Clears the ring buffer and file handle.
        """
        if self.is_open:
            self.close()
        self.file = None
        super().clear()

    def close(self) -> "RingBuffer":
        """
        Closes the file handle and reorganizes the data if the "do_reorganize" attribute is set to True.
        """
        if self.is_open:
            self.write_pointers()
            self.file.close()
            self._is_open = False
            if self.do_reorganize:
                self.reorganize()
                self._remove_raw()

    def open(self) -> "RingBuffer":
        """
        Opens the file handle.
        """
        if not self.is_open:
            self.file = open(self.raw_path, "wb", buffering=0)
            self._is_open = True

    def write_pointers(self) -> "RingBuffer":
        """
        Writes the pointer index to file for reshaping written data
        """
        with open(self.org_path.with_name(self.org_path.name + "_pointers"), "w") as f:
            f.write(str(self.pointer))

    def reorganize(self) -> "RingBuffer":
        """
        Reorganizes the data in the raw file into a single numpy array in the shape defined by the property
        :property:`RingBuffer.data_shape <RF_PROBE.data.RingBuffer.data_shape>`.
        """
        if self.pointer >= 1:
            raw_data = np.memmap(self.raw_path, dtype=self.data_type)
            self._validate_raw(raw_data)
            organized_data = np.memmap(self.org_path,
                                       mode="w+",
                                       shape=self.data_shape,
                                       dtype=self.data_type)
            # generate an index to grab each block of written data
            # org_idx = np.arange(self.data_shape[-1], dtype=np.int64)
            # org_idx = org_idx[0:-1:self.element_length].tolist() if self.element_length > 1 else org_idx.tolist()
            # org_idx.append(None)
            # raw_idx = np.arange(self.data_size, dtype=np.int64)
            # raw_idx = raw_idx[0:-1:self.block_size].tolist() if self.block_size > 1 else raw_idx.tolist()
            # raw_idx.append(None)
            org_idx = range(self.data_shape[-1])
            org_idx = list(org_idx[0:-1:self.element_length]) if self.element_length > 1 else list(org_idx)
            org_idx.append(None)
            raw_idx = range(self.data_size)
            raw_idx = list(raw_idx[0:-1:self.block_size]) if self.block_size > 1 else list(raw_idx)
            raw_idx.append(None)
            for block in range(self.pointer):
                organized_data[..., org_idx[block]:org_idx[block + 1]] \
                    = raw_data[raw_idx[block]:raw_idx[block + 1]].reshape(self.block_shape)
            organized_data.flush()

    def get_data(self) -> np.ndarray:
        """
        Retrieves the n-most recent blocks of data from the ring buffer, where n is the size of the ring buffer
        ("ring_size").

        :returns: The data from the ring buffer as a single numpy array where the rightmost column is the most recent.
        """
        return self.__call__()

    def _initialize_data(self) -> "RingBuffer":
        """
        Initializes the ring buffer with zeros.
        """
        if self.element_depth == 1 and self.element_length == 1:
            null_data = 0
        else:
            null_data = np.zeros(self.block_shape, dtype=self.data_type)
        for _ in range(len(self)):
            self.pop()
            self.appendleft(null_data)

    def _remove_raw(self) -> "RingBuffer":
        """
        Removes the raw data file if performing reorganization.
        """
        self.raw_path.unlink(missing_ok=True)

    def _validate_raw(self, raw_data: np.memmap) -> "RingBuffer":
        """
        Validates the raw data file to ensure that it contains the correct number of elements to be reorganized into
        into the shape defined by :property:`RingBuffer.data_shape <RF_PROBE.data.RingBuffer.data_shape>`.

        :param raw_data: The raw data to be validated.
        :raises DataReorganizationWarning: Raised if the number of elements in the raw data file does not match the
            number of elements required to reorganize the data. In this case, the raw data file will not be deleted.
        """
        try:
            assert len(raw_data) == self.data_size
            _ = raw_data.reshape(self.data_shape)
        except (AssertionError, ValueError):
            warnings.warn(DataReorganizationWarning(raw_data.size, self.data_shape), stacklevel=2)

    def _write_data(self, data: np.ndarray) -> "RingBuffer":
        """
        Writes data to the file handle.

        :param data: The data to be written to the file in the shape of
            :property:`RingBuffer.block_shape <RF_PROBE.data.RingBuffer.block_shape>`
        """
        self.file.write(data)
        # self.file.flush()

    def __call__(self) -> np.ndarray:
        """
        Stacks the data in the ring buffer into a single numpy array where the rightmost column is the most recent.
        """
        return np.hstack(self)


class TimestampedRingBuffer(RingBuffer):
    """
    Fast, fixed-size FIFO thread-safe ring buffer for numpy data with concurrent writing to file. Implemented as
    doubly-linked list for O(1) append and pop operations. A second file is creating with timestamps indicating the
    addition of each block of data to the ring buffer. The timestamps are in microseconds with one decimal place.

    :param element_depth: The number of rows in each block of data
    :param element_length: The number of columns in each block of data
    :param ring_size: The number of blocks of data to store in the ring buffer
    :param path: The path to the file where the data will be written
    :param dtype: The data type of the data to be written
    :param do_reorganize: Whether to reorganize the data into a single numpy array after writing; default is True
    :raises DataReorganizationWarning: Raised if the number of elements in the raw data file does not match the number
     of elements required to reorganize the data. In this case, the raw data file will not be deleted.

    .. note :: This class adheres to the structural requirements of
        :class:`DataBuffer <RF_PROBE.extensions.DataBuffer>`

    .. seealso:: This class inherits from :class:`RingBuffer <RF_PROBE.data.RingBuffer>`

    .. seealso:: This class contains the :class:`Timestamping <RF_PROBE.data.Timestamping>` object for writing
        timestamps
    """
    __slots__ = [
        "pointer",
        # "block_shape", # Turns out that using slots prevents the use of cached_properties
        # "block_size", # I actually think it might be premature optimization to incorporate
        # "data_shape", # a property that calls a cached function, so commenting out for now
        # "data_size", # and leaving the properties as standard
        "element_depth",
        "element_length",
        "raw_path",
        "org_path",
        "data_type",
        "file",
        "timestamps",
        "do_reorganize",
        "_is_open"
    ]

    def __init__(self,
                 element_depth: int,
                 element_length: int,
                 ring_size: int,
                 path: Path,
                 dtype: np.dtype = np.float64,
                 do_reorganize: bool = True):
        """
        Fast, fixed-size FIFO thread-safe ring buffer for numpy data with concurrent writing to file. Implemented as
        doubly-linked list for O(1) append and pop operations. A second file is creating with timestamps indicating the
        addition of each block of data to the ring buffer. The timestamps are in microseconds with one decimal place.

        :param element_depth: The number of rows in each block of data
        :param element_length: The number of columns in each block of data
        :param ring_size: The number of blocks of data to store in the ring buffer
        :param path: The path to the file where the data will be written
        :param dtype: The data type of the data to be written
        :param do_reorganize: Whether to reorganize the data into a single numpy array after writing; default is True
        :raises DataReorganizationWarning: Raised if the number of elements in the raw data file does not match the
        number of elements required to reorganize the data. In this case, the raw data file will not be deleted.

        .. note :: This class adheres to the structural requirements of
            :class:`DataBuffer <RF_PROBE.extensions.DataBuffer>`

        .. seealso:: This class inherits from :class:`RingBuffer <RF_PROBE.data.RingBuffer>`

        .. seealso:: This class contains the :class:`Timestamping <RF_PROBE.data.Timestamping>` object for writing
            timestamps
        """
        self.timestamps = Timestamping(path.with_name(path.name + "_timestamps"), ring_size=ring_size)
        super().__init__(element_depth, element_length, ring_size, path, dtype, do_reorganize)

    def add_data(self, data: np.ndarray) -> "TimestampedRingBuffer":
        """
        Adds data to both the ring buffer and to the data file, and adds a timestamp to the timestamp file.

        :param data: The data to be added.
        """
        super().add_data(data)
        self.timestamps.add_data()

    def clear(self) -> "RingBuffer":
        """
        Clears the ring buffer and file handle.
        """
        if self.is_open != self.timestamps.is_open:
            warnings.warn(DesynchronizedBuffersWarning((self.__class__.__name__,
                                                        self.timestamps.__class__.__name__)),
                          stacklevel=2)
        if self.is_open:
            self.close()
            self.timestamps.close()
        self.file = None
        self.timestamps.clear()
        super().clear()

    def close(self) -> "TimestampedRingBuffer":
        """
        Closes the file handle for both the timestamps and data, and reorganizes the data if the "do_reorganize"
        attribute is set to True.
        """
        if self.is_open != self.timestamps.is_open:
            warnings.warn(DesynchronizedBuffersWarning((self.__class__.__name__,
                                                        self.timestamps.__class__.__name__)),
                          stacklevel=2)
        if self.is_open:
            self.timestamps.close()
            super().close()

    def open(self) -> "TimestampedRingBuffer":
        """
        Opens the file handle for both the timestamps and data.
        """
        if self.is_open != self.timestamps.is_open:
            warnings.warn(DesynchronizedBuffersWarning((self.__class__.__name__,
                                                        self.timestamps.__class__.__name__)),
                          stacklevel=2)
        if not self.is_open:
            self.timestamps.open()
            super().open()

    def get_data(self) -> np.ndarray:
        """
        Retrieves the n-most recent blocks of data from the ring buffer, where n is the size of the ring buffer
        ("ring_size").

        :returns: The data from the ring buffer as a single numpy array where the rightmost column is the most recent.
        """
        return self.__call__()

    def get_timestamps(self) -> np.ndarray:
        """
        Retrieves the n-most recent timestamps from the ring buffer, where n is the size of the ring buffer
        ("ring_size").

        :returns: The timestamps from the ring buffer as a single numpy array where the rightmost column is the
            most recent.
        """
        return self.timestamps.get_data()


class Timestamping(RingBuffer):
    """
    Fast, fixed-size FIFO thread-safe ring buffer for timestamps with concurrent writing to file. Implemented as
    doubly-linked list for O(1) append and pop operations. The timestamps are in microseconds with one decimal place.

    :param path: The path to the file where the data will be written
    :param ring_size: The number of timestamps to store in the ring buffer; default is 1
    :raises DataReorganizationWarning: Raised if the number of elements in the raw data file does not match the number
        of elements required to reorganize the data. In this case, the raw data file will not be deleted.

    .. note :: This class adheres to the structural requirements of
        :class:`DataBuffer <RF_PROBE.extensions.DataBuffer>`

    .. seealso:: This class inherits from :class:`RingBuffer <RF_PROBE.data.RingBuffer>`

    .. seealso:: This class utilizes the :function:`timer <RF_PROBE.timing.timer>` function for collecting
        timestamps.
    """
    __slots__ = [
        "pointer",
        # "block_shape", # Turns out that using slots prevents the use of cached_properties
        # "block_size", # I actually think it might be premature optimization to incorporate
        # "data_shape", # a property that calls a cached function, so commenting out for now
        # "data_size", # and leaving the properties as standard
        "element_depth",
        "element_length",
        "raw_path",
        "org_path",
        "data_type",
        "file",
        "do_reorganize",
        "_is_open"
    ]

    def __init__(self, path: Path, ring_size: int = 1):
        """
        Fast, fixed-size FIFO thread-safe ring buffer for timestamps with concurrent writing to file. Implemented as
        doubly-linked list for O(1) append and pop operations. The timestamps are in microseconds with one decimal
        place.

        :param path: The path to the file where the data will be written
        :param ring_size: The number of timestamps to store in the ring buffer; default is 1
        :raises DataReorganizationWarning: Raised if the number of elements in the raw data file does not match the
        number of elements required to reorganize the data. In this case, the raw data file will not be deleted.

        .. note :: This class adheres to the structural requirements of
            :class:`DataBuffer <RF_PROBE.extensions.DataBuffer>`

        .. seealso:: This class inherits from :class:`RingBuffer <RF_PROBE.data.RingBuffer>`

        .. seealso:: This class utilizes the :function:`timer <RF_PROBE.timing.timer>` function for collecting
            timestamps.
        """
        super().__init__(1, 1, ring_size, path, np.float64, False)

    def add_data(self, data: Any = None) -> "Timestamping":  # noqa: U100
        """
        Adds a timestamp to the file.

        :param data: This argument is here for compatibility with the RingBuffer class, but is not used.

        ..warning:: The data argument is not used in this method and only exists for compatibility with the parent
            :class:`RingBuffer <RF_PROBE.data.RingBuffer>` class.

        ..seealso:: :class:`RingBuffer <RF_PROBE.data.RingBuffer>`
        """
        super().add_data(np.asarray(timer()))



from typing import Union, Tuple, Optional

import numpy as np

from ..configs import HoistedConfig
from ..registries import CallbackRegistry, Provider
from ..requests import AdapterRequest
from ..tools import HAS


@Provider()
class AnalogPulse:
    """
    Higher-level interface for operating linear actuator

    """

    def __init__(self,
                 config: Optional["HoistedConfig"] = None,
                 name: str = "AnalogPulse",
                 device_id: str = "SimDAQ",
                 analog_pulse_channel_key: str = "analog_pulse",
                 ):
        self.config = config
        self.name = config.name if HAS(config) else name
        self.device_id = config.device_id if HAS(config) else device_id
        self.analog_pulse_channel_key = config.analog_pulse_channel_key if HAS(config) else analog_pulse_channel_key
        self.component = AdapterRequest(self,
                                        "AnalogWriter",
                                        alias="component",
                                        wildcard=("device_id", self.device_id))
        self._initialized = False

    @property
    def initialized(self) -> bool:
        return self._initialized

    @initialized.setter
    def initialized(self, value: bool) -> None:
        if value and not self._initialized:
            self._initialized = value

    @property
    def channel_number(self) -> int:
        return self.component.channel_index.get(self.analog_pulse_channel_key)

    @property
    def voltage_range(self) -> Tuple[float, float]:
        return self.component.channel_limits.get(self.analog_pulse_channel_key)

    @CallbackRegistry.register(alias="analog_pulse_command")
    def command(self, value: float) -> None:
        self.put(value)
        self.write()

    @CallbackRegistry.register(alias="analog_pulse_put")
    def put(self, value: float) -> None:
        value = self.convert_percent_to_volts(value, self.voltage_range)
        self.component.data[self.channel_number, :] = value

    @CallbackRegistry.register(alias="analog_pulse_write")
    def write(self) -> None:
        self.component.write(self.component.data, auto_start=True, timeout=self.component.timeout)

    @staticmethod
    def __name__() -> str:
        return "AnalogPulse"



from typing import Mapping, Optional
import warnings

from psychtoolbox.audio import Stream
from psychtoolbox.audio import verbosity as _set_verbosity
from scipy.interpolate import PchipInterpolator

from ..exceptions import SpeakerPropertyError, SpeakerSearchError, SpeakerVolumeWarning, SingletonError
from ..configs import HoistedConfig
from ..tools import HAS, find_devices
from ..timing import Milliseconds, Time, Seconds
from ..registries import Provider

"""
This speaker component inherits the Psychtoolbox audio package to interface with the speaker at low-latencies. Their 
implementation is quite good from my testing, so I don't have to reinvent the wheel write an interface or bindings to 
WASAPI. At least right now... It does mention it's experimental, so maybe I'll have to revisit this later and write my
own wrapper to WASAPI and make python bindings. package doesn't have any particular dependencies (just numpy) so not 
worried about dependency explosion. This is simply the speaker itself and is not intended to be used for stimuli 
generation and/or control. The speaker contains a LUT class that will ensure any stimuli are appropriately intense. 
The speaker is also capable of searching for itself during construction.
"""


_set_verbosity(0)


@Provider()
class Speaker(Stream):
    """
    This is a wrapper around the Psychtoolbox audio.Stream class alongside some complementary methods.
    It is used to connect to the speaker that will be used to delivery auditory stimuli.

    :param config: Config instance that contains a speaker sub-configuration.

    :param properties: A dictionary containing the properties of the speaker. If None, the speaker properties
    will be  located automatically.
    """

    # noinspection PyMissingConstructor
    def __init__(self,
                 config: Optional["HoistedConfig"] = None,
                 properties: Optional[Mapping[str, str]] = None,
                 speaker_id: str = "Digital",
                 lut: Optional[Mapping[str, str]] = None):
        """
        This is a wrapper around the Psychtoolbox audio.Stream class alongside some complementary methods.
        It is used to connect to the speaker that will be used to delivery auditory stimuli.

        :param config: The configuration that contains a hoisted speaker sub-configuration.
        :param properties: A dictionary containing the properties of the speaker. If None, the speaker properties
        will be  located automatically.
        :param speaker_id: The name of the speaker (used only if config is not provided).
        :param lut: A dictionary containing the look-up table for relating sound intensity to speaker volume (used only
            if config is not provided).

        """
        ################################################################################################################
        # REFERENCE CONFIGURATION
        ################################################################################################################
        self.config = config

        ################################################################################################################
        # GENERATE LUT TO RELATE SOUND INTENSITY TO SPEAKER VOLUME
        ################################################################################################################
        self.lut = config.lut if HAS(config) else lut
        self.lut = LUT(self.lut) if HAS(self.lut) else self.lut

        ################################################################################################################
        # FIND AND SET SPEAKER PROPERTIES. MAKE SURE WE ARE USING LOW-LATENCY API.
        ################################################################################################################
        self.speaker_id = config.speaker_id if HAS(config) else speaker_id

        self._properties = properties
        if properties is None:
            self._properties = self.search(name=self.speaker_id)
        self._validate()

        ################################################################################################################
        # INITIALIZE PSYCHTOOLBOX STREAM
        ################################################################################################################
        self.build()

    def __str__(self) -> str:
        return (f"{self.name}: "
                f"latency={self.latency.milliseconds} ms, "
                f"channels={self.channels}, "
                f"sampling rate={self.sampling_rate} Hz")

    @property
    def device_index(self) -> int:
        return int(self._properties.get("DeviceIndex"))

    @property
    def latency(self) -> "Time":
        return Seconds(self._properties.get("HighOutputLatency"))

    @property
    def name(self) -> str:
        return self._properties.get("DeviceName")

    @property
    def channels(self) -> int:
        return int(self._properties.get("NrOutputChannels"))

    @property
    def sampling_rate(self) -> int:
        return int(self._properties.get("DefaultSampleRate"))

    @property
    def start_time(self) -> float:
        return self.status.get("StartTime") if self.running else None

    @property
    def running(self) -> bool:
        return True if self.status.get("Active") == 1 else False

    @classmethod
    def search(cls, name: Optional[str] = None) -> Mapping:
        speakers = find_devices("speaker")
        for speaker_name, properties in speakers.items():
            if name is None:
                return properties
            elif name.lower() in speaker_name.lower():
                return properties
            else:  # pragma: no cover
                pass
        raise SpeakerSearchError(name, speakers)  # pragma: no cover

    def build(self) -> "Speaker":
        return Stream.__init__(self,
                               device_id=self.device_index,
                               latency_class=3,
                               freq=self.sampling_rate,
                               channels=self.channels,
                               mode=9)

    def start(self, *args, **kwargs):
        Stream.start(self, 0, 0, 0)

    def stop(self, *args, **kwargs):
        Stream.stop(self)

    def destroy(self):
        Stream.close(self)

    def __del__(self):
        self.destroy()

    def _validate(self):
        # make sure exists
        located_speakers = find_devices("speaker")

        if not any([speaker_name for speaker_name in located_speakers.keys()
                    if self.name in speaker_name]):  # pragma: no cover
            raise SpeakerSearchError(self.name, located_speakers)

        # make sure is WASAPI
        if self._properties.get("HostAudioAPIName") != "Windows WASAPI":  # pragma: no cover
            raise SpeakerPropertyError(self.name,
                                       "HostAudioAPIName",
                                       self._properties.get("HostAudioAPIName"),
                                       "Windows WASAPI")


class LUT(PchipInterpolator):
    """
    Look-up table relating sound intensity (decibels) to normalized speaker volume. This is used to ensure that the
    audio is appropriately intense. The LUT is interpolated using piecewise cubic hermite interpolating polynomial
    (PCHIP).

    """
    def __init__(self, lut_map: Mapping):
        self.lut_map = lut_map
        self.lut_map = {float(key): value for key, value in self.lut_map.items()}
        self._x = list(self.lut_map.keys())
        self._y = list(self.lut_map.values())
        self.rescale()
        super().__init__(self._x, self._y)

    def get(self, x: float) -> float:

        if x < min(self._x) or x > max(self._x):
            # return min or max volume if outside of range (whichever is closer)
            closer_to_max = abs(max(self._x) - x) < abs(min(self._x) - x)
            if closer_to_max:
                set_volume = 1.0
            else:
                set_volume = 0.0
            warnings.warn(SpeakerVolumeWarning(x, min(self._x), max(self._x), set_volume))
            return set_volume
        else:
            return super().__call__(x)

    def rescale(self):
        if not any([0 <= y <= 1 for y in self._y]):
            self._y = [y / max(self._y) for y in self._y]



from .behavior import Behavior


# Autoregistration
from ..extensions import BehaviorState
from ..registries import auto_register, BehaviorRegistry
from . import states
from . import machine
from . import transitions

__modules__ = (machine, transitions)

#for module_ in __modules__:
#    auto_register(module_, BehaviorRegistry)

# Exposure
__all__ = [
    "Behavior",
]



import pytest
from neurobeam.registries import (AdapterRegistry, AudioRegistry, BehaviorRegistry, CallbackRegistry, ComponentRegistry,
                                  ProviderRegistry, auto_register)
from neurobeam.exceptions import (InvalidAdapterError, InvalidComponentError, InvalidCallbackError, InvalidStateError,
                                  InvalidAudioError)


def test_adapter_registry(MockReader, MockWriter, MockTrigger, helper):
    # Test that the AdapterRegistry is registering the correct classes
    AdapterRegistry.register()(MockReader)
    assert AdapterRegistry.has("MockReader_")
    AdapterRegistry.register()(MockWriter)
    assert AdapterRegistry.has("MockWriter_")
    AdapterRegistry.register()(MockTrigger)
    assert AdapterRegistry.has("MockTrigger_")

    # Test that typecheck works
    #assert AdapterRegistry.type_check(MockReader, raise_exception=True)

    # Test that the AdapterRegistry is returning the correct classes
    assert AdapterRegistry.get("MockReader_") == MockReader
    assert AdapterRegistry.get("MockWriter_") == MockWriter
    assert AdapterRegistry.get("MockTrigger_") == MockTrigger

    # Test that the AdapterRegistry won't register random classes
    RandomClass = helper.empty_class("RandomClass")
    with pytest.raises(InvalidAdapterError):
        AdapterRegistry.type_check(RandomClass, raise_exception=True)
    AdapterRegistry.register()(RandomClass)
    assert not AdapterRegistry.has("RandomClass")

    # Test Approximate Retrieval
    assert AdapterRegistry.get("MockRead", approximate=True) == MockReader

    # Test Alias Retrieval
    assert AdapterRegistry.register(alias="MR")(MockReader)
    assert AdapterRegistry.has("MR")


def test_audio_registry(helper):

    def mock_audio(*args, **kwargs):
        return

    # Test that the AudioRegistry is registering the correct classes
    AudioRegistry.register()(mock_audio)
    assert AudioRegistry.has("mock_audio")

    # Test that typecheck works
    assert AudioRegistry.type_check(mock_audio, raise_exception=True)

    # Test that the AudioRegistry is returning the correct classes
    assert AudioRegistry.get("mock_audio") == mock_audio

    # Test that the AudioRegistry won't register random classes
    RandomClass = helper.empty_class("RandomClass")
    with pytest.raises(InvalidAudioError):
        AudioRegistry.type_check(RandomClass, raise_exception=True)
    AudioRegistry.register()(RandomClass)
    assert not AudioRegistry.has("RandomClass")

    # Test Approximate Retrieval
    assert AudioRegistry.get("mock_audi", approximate=True) == mock_audio

    # Test Alias Retrieval
    assert AudioRegistry.register(alias="MA")(mock_audio)
    assert AudioRegistry.has("MA")


def test_behavior_registry(MockState, helper):
    # Test that the BehaviorRegistry is registering the correct classes
    BehaviorRegistry.register()(MockState)
    assert BehaviorRegistry.has("MockState_")

    # Test that typecheck works
    #assert BehaviorRegistry.type_check(MockState, raise_exception=True)

    # Test that the BehaviorRegistry is returning the correct classes
    assert BehaviorRegistry.get("MockState_") == MockState

    # Test that the BehaviorRegistry won't register random classes
    RandomClass = helper.empty_class("RandomClass")
    with pytest.raises(InvalidStateError):
        BehaviorRegistry.type_check(RandomClass, raise_exception=True)
    BehaviorRegistry.register()(RandomClass)
    assert not BehaviorRegistry.has("RandomClass")

    # Test Approximate Retrieval
    assert BehaviorRegistry.get("MockSta", approximate=True) == MockState

    # Test Alias Retrieval
    assert BehaviorRegistry.register(alias="MS")(MockState)
    assert BehaviorRegistry.has("MS")


def test_callback_registry(helper):
    mock_callback = helper.mock_callback()
    true_name = mock_callback.__name__

    # Test that the CallbackRegistry is registering the correct callbacks
    CallbackRegistry.register()(mock_callback)
    assert CallbackRegistry.has(true_name)

    # Test that typecheck works
    assert CallbackRegistry.type_check(mock_callback, raise_exception=True)

    # Test that the CallbackRegistry is returning the correct callbacks
    assert CallbackRegistry.get(true_name) == mock_callback

    # Test that the CallbackRegistry won't register random classes
    RandomCallback = helper.empty_class("RandomCallback")
    with pytest.raises(InvalidCallbackError):
        CallbackRegistry.type_check(RandomCallback, raise_exception=True)
    CallbackRegistry.register()(RandomCallback)
    assert not CallbackRegistry.has("RandomCallback")

    # Test Approximate Retrieval
    assert CallbackRegistry.get(true_name[:5], approximate=True) == mock_callback

    # Test Alias Retrieval
    assert CallbackRegistry.register(alias="MC")(mock_callback)
    assert CallbackRegistry.has("MC")

    # Test collection retrieval
    assert len(CallbackRegistry.get_collection("linear_actuator")) == 3


def test_component_registry(MockComponent, helper):
    # Test that the ComponentRegistry is registering the correct classes
    ComponentRegistry.register()(MockComponent)
    assert ComponentRegistry.has("MockComponent_")

    # Test that typecheck works
    assert ComponentRegistry.type_check(MockComponent, raise_exception=True)

    # Test that the ComponentRegistry is returning the correct classes
    assert ComponentRegistry.get("MockComponent_") == MockComponent

    # Test that the ComponentRegistry won't register random classes
    RandomClass = helper.empty_class("RandomClass")
    with pytest.raises(InvalidComponentError):
        ComponentRegistry.type_check(RandomClass, raise_exception=True)
    ComponentRegistry.register()(RandomClass)
    assert not ComponentRegistry.has("RandomClass")

    # Test Approximate Retrieval
    assert ComponentRegistry.get("MockCompon", approximate=True) == MockComponent

    # Test Alias Retrieval
    assert ComponentRegistry.register(alias="MC")(MockComponent)
    assert ComponentRegistry.has("MC")


def test_provider_registry(MockComponent):
    ProviderRegistry.register(MockComponent)
    mock_component = MockComponent()
    ProviderRegistry.has("MockComponent")


def test_request_registry():
    ...


def test_auto_register():
    ...



from typing import Tuple  # Optional
from functools import partial
from weakref import proxy

import numpy as np
from psychtoolbox.audio import Slave

from ..configs import HoistedConfig
from ..hardware.speaker import Speaker
from ..factories import AudioFactory
from ..requests import AdapterRequest
from ..registries import CallbackRegistry, Provider
from ..timing import Time, Seconds
from ..tools import HAS


@Provider()
class AudioStimulus(Slave):
    # noinspection PyMissingConstructor
    def __init__(self,
                 config: "HoistedConfig",
                 name: str = "audio_stimulus",
                 style: str = "white_noise",
                 duration: Time = Seconds(1),
                 options=None,
                 stimulus: np.ndarray = None,
                 device_id: str = "Digital",
                 speaker: "Speaker" = None
                 ):
        ################################################################################################################
        # REPORTING PROGRESS
        ################################################################################################################
        self.config = config
        self.name = config.name if HAS(config) else name
        #: int: The number of times the cue has been triggered.
        self._triggers = 0
        #: int: The number of times the cue has been stopped.
        self._stops = 0
        #: np.ndarray: The stimulus to be presented.
        self._stimulus = stimulus
        self._options = config.options if HAS(config) and HAS(config.options) else {}
        self._duration = config.duration if HAS(config) else duration
        self._style = config.style if HAS(config) else style
        self.component = proxy(speaker) if HAS(speaker) else AdapterRequest(self,
                                                                            "Speaker",
                                                                            alias="component",
                                                                            wildcard=("speaker_id",
                                                                                      self.config.device_id))
        self.device_id = config.device_id if HAS(config) else device_id

        ################################################################################################################
        # INITIALIZATION
        ################################################################################################################
        #Slave.__init__(self, speaker.handle, data=stimulus, channels=speaker.channels)
        self._initialized = False

    @property
    def initialized(self) -> bool:
        return self._initialized

    @initialized.setter
    def initialized(self, value: bool) -> None:
        if value and not self._initialized:
            factory = AudioFactory(self.component)
            self._stimulus = factory.create(self._style, **self._options)
            Slave.__init__(self, self.component.handle, data=self._stimulus, channels=self.component.channels)
            self._initialized = value

    @property
    def latency(self) -> float:
        return self.component.latency

    @property
    def running(self) -> bool:
        return self._triggers > self._stops

    @CallbackRegistry.register(alias="audio_put")
    def put(self) -> "AudioStimulus":
        self.fill_buffer(self._stimulus)

    def set_volume(self, volume: float) -> "AudioStimulus":
        self.setVolume(volume)

    @CallbackRegistry.register(alias="audio_command")
    def command(self) -> "AudioStimulus":
        self.put()
        self.write()

    @CallbackRegistry.register(alias="audio_write")
    def write(self) -> "AudioStimulus":
        self.start(1, 0, 0)
        self._triggers += 1



import weakref
from functools import partial
from typing import Tuple, Any, Generator, Iterable, List
from weakref import WeakValueDictionary
import warnings
import inspect

from frozendict import frozendict

# noinspection PyProtectedMember
from .behavior.machine import Machine
from .behavior.states import State
from .exceptions import (MissingIdentifierError,
                         NoWeakRefError,
                         ItemNotAddedWarning)
from .extensions import Component, Adapter
from .hardware.daq import StreamingComponent, OnDemandComponent
from .registries import CallbackRegistry, ProviderRegistry, RequestRegistry
from .requests import AdapterRequest, CallbackRequest
from .tools import EMPTY


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DI CONTAINER
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class Container:
    """
    The Container class is a simple container for holding neurobeam components, adapters, and controllers. It is used to
    during the collection and injection of callbacks into the instances contained within it.
    """

    def __init__(self, *args):
        self.components = WeakValueDictionary()
        self.adapters = WeakValueDictionary()
        self.controller = None

        for arg in args:
            self.add_object(arg)

    def contents(self) -> Tuple[Any, ...]:
        """
        Get a tuple of the objects contained in the container

        :returns: A tuple of the objects contained in the container
        """
        return tuple(self.components.values()) + tuple(self.adapters.values()) + (self.controller(),)

    def add_object(self, obj: Any) -> None:
        """
        Add an object to the container

        :param obj: The object to add
        """
        # to resolve unavoidable circular dependency
        import neurobeam.controller
        try:
            if isinstance(obj, Component):
                self.components[id(obj)] = obj
            elif isinstance(obj, Adapter):
                self.adapters[id(obj)] = obj
            elif isinstance(obj, neurobeam.controller.Controller):
                self.controller = weakref.ref(obj)
            else:
                warnings.warn(ItemNotAddedWarning(obj), stacklevel=2)
        except TypeError:
            raise NoWeakRefError(obj)


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INJECTION
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


def inject(*args) -> None:
    # if we only passed a controller, then extract all the components and adapters
    if len(args) == 1:
        args = (args[0], *args[0].components, *args[0].adapters)

    # collect all the neurobeam components, adapters, and controllers into a container
    container = Container(*args)

    # retrieve the cache of instantiated providers
    providers = ProviderRegistry()

    # retrieve the cache of instantiated callback requests
    requests = RequestRegistry.traverse()

    # bring in the callback registry
    callback_registry = CallbackRegistry()

    # iterate through all requests
    for request in requests:
        if isinstance(request, CallbackRequest):
            request.inject(callback_registry, providers)
        elif isinstance(request, AdapterRequest):
            request.inject(providers)



from typing import Optional

from ..configs import HoistedConfig
from ..registries import Provider
from ..tools import HAS


@Provider()
class SLM:
    """
    SLM hardware class
    """

    def __init__(self,
                 config: Optional["HoistedConfig"] = None,
                 brand: str = "Meadowlark",
                 realtime_mode: bool = False):
        """
        SLM hardware class

        :param config: The hoisted configuration
        :param brand: The brand of the SLM (used only if config is not provided)
        :param realtime_mode: Whether the SLM is in realtime mode (used only if config is not provided)
        """
        #: Optional[HoistedConfig]: The hoisted configuration
        self.config = config
        #: str: The brand of the SLM
        self.brand = config.brand if HAS(config) else brand
        #: bool: The realtime mode of the SLM
        self.realtime_mode = config.realtime_mode if HAS(config) else realtime_mode

    def __str__(self) -> str:
        return f"SLM: {self.brand}, Realtime Mode: {self.realtime_mode}"

    def build(self) -> None:
        """
        Build the SLM
        """
        pass

    def start(self) -> None:
        """
        Start the SLM
        """
        pass

    def stop(self) -> None:
        """
        Stop the SLM
        """
        pass

    def destroy(self) -> None:
        """
        Destroy the SLM
        """
        pass



from ..tools import FORMAT_TERMINAL


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// REPORTING CALLBACKS (CONSOLE)
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


def reporting_console_start() -> None:
    print(FORMAT_TERMINAL("PROBE STARTED", "header"))


def reporting_console_stop() -> None:
    print(FORMAT_TERMINAL("PROBE STOPPED", "header"))


def reporting_console_task_start() -> None:
    print(FORMAT_TERMINAL("TASK STARTED", "header"))


def reporting_console_task_finished() -> None:
    print(FORMAT_TERMINAL("TASK FINISHED", "header"))


def reporting_console_concluded() -> None:
    print(FORMAT_TERMINAL("PROBE CONCLUDED", "header"))


def reporting_console_progress_state(state: str, progress: float) -> None:
    print(FORMAT_TERMINAL(f"{state} Progress: ", "local_progress") + f"{progress:.0f}%")


def reporting_console_progress_task(task: str, progress: float) -> None:
    print(FORMAT_TERMINAL(f"{task.capitalize()} Progress: ", "global_progress") + f"{progress:.0f}%")



from typing import Any
import weakref


"""
Some simple functions to make code more readable. Probably not ideal to use most of these in performance 
critical sections.
"""


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SENTINELS OR SENTINEL-LIKE FUNCTIONS FOR READABILITY
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


# noinspection PyPep8Naming
def HAS(thing: Any) -> bool:
    """
    Check if a thing is not None, not an empty string, not an empty tuple, not an empty list, and not an empty dict.

    :param thing: The thing to check
    :returns: True if the thing is not None, not an empty string, not an empty tuple, not an empty list,
        and not an empty dict.

    .. seealso:: :function:`EMPTY <RF_PROBE.tools.readability.EMPTY>`
    """

    return thing is not None and thing != '' and thing != () and thing != [] and thing != {}


# noinspection PyPep8Naming
def EMPTY(thing: Any) -> bool:
    """
    Check if a thing is None, an empty string, an empty tuple, an empty list, or an empty dict.

    :param thing: The thing to check
    :returns: True if the thing is None, an empty string, an empty tuple, an empty list, or an empty dict.

    .. seealso:: :function:`HAS <RF_PROBE.tools.readability.HAS>`
    """
    return thing is None or thing == '' or thing == () or thing == [] or thing == {}


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DICTIONARY FUNCTIONS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


def parse_empties(mapping: dict) -> dict:
    """
    Parse a dictionary and return a new dictionary with only the non-empty values.

    :param mapping: The dictionary to parse
    :returns: A new dictionary with only the non-empty values.
    """
    return {key: value for key, value in mapping.items() if HAS(value)}


def prune_keys(mapping: dict, keys: list) -> dict:
    """
    Prune a dictionary of specific keys.

    :param mapping: The dictionary to prune
    :param keys: The keys to remove
    :returns: A new dictionary with the specified keys removed
    """
    return {key: value for key, value in mapping.items() if key not in keys}



from . import console
from . import debug

__modules__ = (console, debug)


# Register all callbacks automatically
from ..registries import auto_register as _auto_register, CallbackRegistry
from ..extensions import CallbackFunction


for module_ in __modules__:
    _auto_register(module_, CallbackRegistry)



from .linear_actuator import LinearActuator
from .analog_pulse import AnalogPulse
from .audio_stimulus import AudioStimulus
from .stimulus_trigger import StimulusTrigger
from .digital_trigger import DigitalTrigger
from typing import Any

# auto-register all the classes in these modules
from ..registries import auto_register as _auto_register
from ..registries import AdapterRegistry
from . import linear_actuator
from . import analog_pulse
from . import audio_stimulus
from . import digital_trigger
from . import stimulus_trigger


__modules__ = (linear_actuator, analog_pulse, audio_stimulus, stimulus_trigger, digital_trigger)

for module_ in __modules__:
    _auto_register(module_, AdapterRegistry)

# Exposure
__all__ = [
    "AudioStimulus"
    "AnalogPulse",
    "DigitalTrigger",
    "LinearActuator",
    "StimulusTrigger",
]



from .configs import Config
from .controller import Controller
from .factories import AdapterFactory, ComponentFactory
from .tools import HAS


"""
This function instantiates a populated neurobeam controller from the specified configuration. It should be used in the
program's main loop.
"""


def create_controller(config: "Config") -> "Controller":
    """
    Create a controller populated with adapters and components from the given configuration.

    :param config: The configuration of neurobeam used to create the controller
    :returns: The instantiated controller

    .. seealso:: :class:`Config <neurobeam.configs.Config>`\, :class:`Controller <neurobeam.controller.Controller>`
    """
    if HAS(config.components):
        components = (ComponentFactory(config).create(component) for component in range(len(config.components)))
    else:
        components = iter(())

    if HAS(config.adapters):
        adapters = (AdapterFactory(config).create(adapter) for adapter in range(len(config.adapters)))
    else:
        adapters = iter(())

    return Controller(*components, *adapters)



from typing import Sequence
from random import shuffle, sample


def random_shuffle(sequence: Sequence) -> tuple:
    """
    Randomly shuffle a sequence.

    :param sequence: The sequence to shuffle.
    :returns: A tuple containing the shuffled sequence.
    """
    sequence = list(sequence)
    shuffle(sequence)
    return tuple(sequence)


def balanced_shuffle(sequence: Sequence) -> tuple:
    """
    Counter-balanced shuffling of some sequence

    :param sequence: The sequence to shuffle.
    :returns: A tuple containing the shuffled sequence.
    """
    sequence = list(sequence)
    length = len(sequence)
    unique_values = list(set(sequence))
    iters_per_value = [sequence.count(value) for value in unique_values]
    if not all((iters == iters_per_value[0] for iters in iters_per_value)):
        raise ValueError("The sequence is not balanced.")
    n_blocks = iters_per_value[0]
    block = unique_values
    blocks = [block for _ in range(n_blocks)]
    shuffled_sequence = []
    for _ in range(n_blocks):
        shuffle(block)
        shuffled_sequence.extend(block)
    return tuple(shuffled_sequence)



from typing import Union
import asyncio
import subprocess
from enum import IntEnum
from collections import ChainMap

from winsdk.windows.devices.enumeration import DeviceInformation
from winsdk.windows.media.devices import VideoDeviceController
from frozendict import frozendict
from nidaqmx import system
from psychtoolbox import audio


"""
Tools for locating various devices.
"""


class DeviceClass(IntEnum):
    """
    DeviceClass is an enumeration of the different classes of devices utilized by PROBE.

    The SPEAKER and CAMERA enumerated constants reflect their enumerations in the Windows Runtime API
    'DeviceClass <https://learn.microsoft.com/en-us/uwp/api/windows.devices.enumeration.deviceclass?view=winrt-22621>'.
    This enumeration is outlined below. There is a python-interface for WinRT, but it is unmaintained. The python
    package winsdk is a community-driven fork of WinRT that seems suitable, albeit lightly documented.

    Probe enumerated device class constants
    SPEAKER = 2
    CAMERA = 4
    DAQ = 7
    ALL = 8

    Windows Runtime API DeviceClass Enum
    All = 0
    AudioCapture = 1
    AudioRender = 2
    PortableStorageDevice = 3
    VideoCapture = 4
    ImageScanner = 5
    Location = 6

    """
    SPEAKER = 2
    CAMERA = 4
    DAQ = 7
    ALL = 8

    @classmethod
    def from_string(cls, device_class: str):
        device_class = device_class.upper()
        if device_class in cls.__members__.keys():
            return getattr(cls, device_class)
        else:
            raise ValueError

    @classmethod
    def from_arg(cls, device_class: Union[int, str]):
        if isinstance(device_class, str):
            return cls.from_string(device_class)
        elif isinstance(device_class, int):
            return cls(device_class)
        else:
            raise ValueError


def find_devices(device_class: Union[int, str, "DeviceClass"]) -> Union[frozendict, ChainMap]:
    """
    Finds devices associated with the specified device-class enumerated constant.

    """
    # convert to enumerated constant if int or str
    device_class = DeviceClass.from_arg(device_class)

    # find devices for each type
    if device_class == DeviceClass.SPEAKER:
        return _find_speakers()
    elif device_class == DeviceClass.CAMERA:
        return _find_cameras()
    elif device_class == DeviceClass.DAQ:
        devices = _get_daqs()
    elif device_class == DeviceClass.ALL:
        # get all enumerated device constants except for All, then recursively call find_devices()
        enums = list(DeviceClass.__members__.values())[:-1]
        return ChainMap([find_devices(device_class_) for device_class_ in enums])
    else:
        raise ValueError


def _find_cameras():
    """
    Find cameras using WinRT

    """
    devices = asyncio.run(_get_winrt_devices(DeviceClass.CAMERA))
    return frozendict({device.name: (idx, device) for idx, device in enumerate(devices)})


def _find_speakers() -> frozendict:
    """
    Find speakers using the low-latency Windows Audio Session API (WASAPI) through combination of WinRT &
    psychtoolbox's PsychPortAudio.

    """
    # get speakers using winRT
    devices = asyncio.run(_get_winrt_devices(DeviceClass.SPEAKER))
    # make mapping
    devices = {device.name: None for device in devices}
    # pair properties of speaker , device_type 13 flags WASAPI only
    properties = audio.get_devices(device_type=13)
    names = [device.get("DeviceName") for device in properties]
    for device in devices.keys():
        idx = [position for position, string in enumerate(names) if string.lower() in device.lower()]
        devices[device] = properties[idx[-1]]   # If audio.get_devices erroneously returns high and
        # low-latency implementations, the low-latency implementation will be second
    return frozendict(devices)


async def _get_winrt_devices(device_class: int):
    return await DeviceInformation.find_all_async(device_class)


def _get_daqs():
    devices = system.System.local().devices
    return frozendict({device.name: device.product_type for device in devices})



from typing import Union, Tuple, Optional

from memoization import cached
import numpy as np

from ..configs import HoistedConfig
from ..registries import CallbackRegistry, Provider
from ..requests import AdapterRequest
from ..timing import timer
from ..tools import HAS

from tests.conftest import report_function_call


@Provider()
class LinearActuator:
    """
    Higher-level interface for operating linear actuator

    """
    
    def __init__(self, 
                 config: Optional["HoistedConfig"] = None,
                 name: str = "LinearActuator",
                 device_id: str = "SimDAQ",
                 command_channel_key: str = "linear_actuator_command",
                 physical_range: Tuple[float, float] = (0.0, 50.0),
                 physical_units: str = "mm",
                 ):
        self.config = config
        self.name = config.name if HAS(config) else name
        self.device_id = config.device_id if HAS(config) else device_id
        self.command_channel_key = config.command_channel_key if HAS(config) else command_channel_key
        self.physical_range = config.physical_range if HAS(config) else physical_range
        self.physical_units = config.physical_units if HAS(config) else physical_units
        self.component = AdapterRequest(self,
                                        "AnalogWriter",
                                        alias="component",
                                        wildcard=("device_id", self.device_id))
        self._initialized = False

    @property
    def initialized(self) -> bool:
        return self._initialized

    @initialized.setter
    def initialized(self, value: bool) -> None:
        if value and not self._initialized:
            self._initialized = value

    @staticmethod
    @cached(max_size=None)
    def convert_percent_to_volts(percent: np.ndarray, voltage_range: Tuple[float, float]) -> Union[float, np.ndarray]:
        return (percent / 100.0) * (voltage_range[1] - voltage_range[0]) + voltage_range[0]

    @property
    def channel_number(self) -> int:
        return self.component.channel_index.get(self.command_channel_key)

    @property
    def voltage_range(self) -> Tuple[float, float]:
        return self.component.channel_limits.get(self.command_channel_key)

    @CallbackRegistry.register(alias="linear_actuator_command")
    def command(self, value: float) -> None:
        self.put(value)
        self.write()

    @CallbackRegistry.register(alias="linear_actuator_put")
    def put(self, value: float) -> None:
        value = self.convert_percent_to_volts(value, self.voltage_range)
        self.component.data[self.channel_number, :] = value

    @CallbackRegistry.register(alias="linear_actuator_write")
    def write(self) -> None:
        self.component.write(self.component.data, auto_start=True, timeout=self.component.timeout)

    @staticmethod
    def __name__() -> str:
        return "LinearActuator"



from typing import Tuple  # Optional
from functools import partial

import numpy as np

from ..registries import AudioRegistry
from ..timing import Time
from .readability import HAS


@AudioRegistry.register()
def paired_pulse(sampling_rate: int,
                 channels: int,
                 duration: Tuple[Time, Time, Time],
                 methods: Tuple[str, str] = ("white_noise", "white_noise"),
                 volume: Tuple[float, float] = tuple([0.5, 1.0])
                 ) -> np.ndarray:
    # pre-pulse
    # noinspection PyProtectedMember
    pre_pulse = AudioRegistry._registry.get(methods[0])(sampling_rate, channels, duration[0], volume[0])

    # pulse gap
    # noinspection PyProtectedMember
    pulse_gap = AudioRegistry._registry.get("silence")(sampling_rate, channels, duration[1])

    # post-pulse
    # noinspection PyProtectedMember
    post_pulse = AudioRegistry._registry.get(methods[1])(sampling_rate, channels, duration[2], volume[1])

    samples = np.concatenate((pre_pulse, pulse_gap, post_pulse))

    return samples


@AudioRegistry.register()
def sawtooth(sampling_rate: int, channels: int, duration: Time, frequency: float) -> np.ndarray:
    duration = duration.seconds
    samples = int(sampling_rate * duration)
    time = np.arange(samples) / sampling_rate
    samples = 2 * (frequency * time - np.floor(frequency * time + 0.5))
    samples = samples.clip(-1, 1)
    if channels > 1:
        samples = np.tile(samples, (channels, 1)).T
    return samples


@AudioRegistry.register()
def silence(sampling_rate: int, channels: int, duration: Time) -> np.ndarray:
    duration = duration.seconds
    samples = int(sampling_rate * duration)
    return np.zeros((samples, channels))


@AudioRegistry.register()
def sine(sampling_rate: int, channels: int, duration: Time, frequency: float, volume = 1) -> np.ndarray:
    duration = duration.seconds
    samples = int(sampling_rate * duration)
    time = np.arange(samples) / sampling_rate
    samples = np.sin(2 * np.pi * frequency * time)
    if channels > 1:
        samples = np.tile(samples, (channels, 1)).T
    samples = rescale(samples, volume)
    return samples


@AudioRegistry.register()
def square(sampling_rate: int, channels: int, duration: Time, frequency: float) -> np.ndarray:
    duration = duration.seconds
    samples = int(sampling_rate * duration)
    time = np.arange(samples) / sampling_rate
    samples = np.sign(np.sin(2 * np.pi * frequency * time))
    if channels > 1:
        samples = np.tile(samples, (channels, 1)).T
    return samples


@AudioRegistry.register()
def triangle(sampling_rate: int, channels: int, duration: Time, frequency: float) -> np.ndarray:
    duration = duration.seconds
    samples = int(sampling_rate * duration)
    time = np.arange(samples) / sampling_rate
    samples = 2 * np.abs(2 * (time * frequency - np.floor(time * frequency + 0.5)))
    samples -= 1
    if channels > 1:
        samples = np.tile(samples, (channels, 1)).T
    return samples


@AudioRegistry.register()
def white_noise(sampling_rate: int, channels: int, duration: Time, volume: float = 1.0) -> np.ndarray:
    duration = duration.seconds
    samples = int(sampling_rate * duration)
    samples = np.random.randn(samples).flatten().T
    samples = samples.clip(-1, 1)
    if channels > 1:
        samples = np.tile(samples, (channels, 1)).T
    samples = rescale(samples, volume)
    return samples


def rescale(samples: np.ndarray, target_volume: float) -> np.ndarray:
    full_scale = [-1, 1]
    new_scale = [-target_volume, target_volume]
    return ((samples - full_scale[0]) * (new_scale[1] - new_scale[0])) / (full_scale[1] - full_scale[0]) + new_scale[0]



from typing import Tuple  # Optional
from weakref import proxy

from ..configs import HoistedConfig
from ..requests import AdapterRequest
from ..registries import Provider
from ..tools import HAS


@Provider()
class DigitalTrigger:
    def __init__(self,
                 config: "HoistedConfig" = None,
                 name: str = "Digital Trigger",
                 transition_key: str = "early_ingress",
                 device_id: str = "SimDAQ",
                 digital_trigger_channel_key: str = "burrow_sensor",
                 component: "DigitalStimulus" = None,
                 ):
        self.config = config
        self.name = config.name if HAS(config) else name
        self.device_id = config.device_id if HAS(config) else device_id
        self.digital_trigger_channel_key = config.digital_trigger_channel_key if HAS(config) \
            else digital_trigger_channel_key
        self.component = proxy(component) if component else (
            AdapterRequest(self, "DigitalMultiInputStream", alias="component", wildcard=("device_id",
                                                                                         self.config.device_id)))
        self.transitions = AdapterRequest(self, "Machine", alias="transitions")
        self._running = False
        self._initialized = False

    @property
    def initialized(self) -> bool:
        return self._initialized

    @initialized.setter
    def initialized(self, value: bool) -> None:
        if value and not self._initialized:
            self._initialized = value

    @property
    def channel_number(self) -> int:
        return self.component.channel_index.get(self.digital_trigger_channel_key)

    @property
    def running(self):
        return self._running

    @running.setter
    def running(self, value):
        self._running = value

    def command(self):
        self.transitions.get("early_ingress")()

    def initialize(self):
        ...

    def read(self):
        return self.component.data[self.channel_number, :]

    def trigger(self, time):
        data = self.read()
        if np.max(data) >= 1:
            self.command()

    def stop(self):
        self.running = False

    def __call__(self):
        self.command()



import warnings
from typing import Any, Optional
from multiprocessing import Process, Queue, Pipe
from pathlib import Path
from operator import eq
from time import sleep

import cv2 as cv
import numpy as np

from ..configs import HoistedConfig
from ..data import TimestampedRingBuffer
from ..exceptions import DualUserInterfaceError
from ..registries import Provider
from ..tools import PatternMatching, HAS


def stream_video(camera_idx: int = 0,
                 height: int = 480,
                 width: int = 640,
                 quarter_period: float = 8.33,
                 display_name: Optional[str] = None,
                 image_queue: Optional[Queue] = None,
                 command_pipe: Optional[Pipe] = None,
                 output_file: Optional[Path] = None
                 ) -> None:

    if output_file:
        ring_buffer = TimestampedRingBuffer(element_depth=height,
                                            element_length=width,
                                            ring_size=1,
                                            path=output_file,
                                            dtype=np.uint8,
                                            do_reorganize=False)
    else:
        ring_buffer = None

    # XOR VALIDATION - MUST HAVE DISPLAY OR GUI
    if not (bool(display_name) ^ bool(image_queue)):  # pragma: no cover
        raise DualUserInterfaceError

    # MAKE WINDOW FOR DISPLAY NAME
    if display_name:
        cv.namedWindow(display_name, cv.WINDOW_NORMAL)

    # CREATE CAMERA FEED
    cam = cv.VideoCapture(camera_idx, cv.CAP_DSHOW)
    #cam = cv.VideoCapture(camera_idx, cv.CAP_MSMF)
    cam.set(cv.CAP_PROP_FRAME_WIDTH, width)
    cam.set(cv.CAP_PROP_FRAME_HEIGHT, height)
    cam.set(cv.CAP_PROP_FOURCC, cv.VideoWriter_fourcc('M', 'J', 'P', 'G'))

    # RUN CAMERA LOOP
    while True:
        # Ensure camera is connected & attempt to read frame
        if cam.isOpened():
            (status, frame) = cam.read()
            if frame.ndim > 2:  # necessary evil to convert to grayscale
                frame = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
            # Do whatever user wishes with the video frame
            with PatternMatching([bool(status), bool(image_queue), bool(ring_buffer)],
                                 [eq, eq, eq]) as case:
                if case([True, True, True]):
                    image_queue.put_nowait(frame)
                    ring_buffer.add_data(frame)
                elif case([True, True, False]):
                    image_queue.put_nowait(frame)
                elif case([True, False, True]):
                    cv.imshow(display_name, frame)
                    ring_buffer.add_data(frame)
                    cv.waitKey(quarter_period)
                elif case([True, False, False]):
                    cv.imshow(display_name, frame)
                    cv.waitKey(quarter_period)
                else:  # pragma: no cover
                    warnings.warn(CameraCollectionWarning(camera_idx))

                if command_pipe.poll():
                    command = command_pipe.recv()
                    if command == "stop":
                        cam.release()
        else:
            if ring_buffer:
                ring_buffer.close()
            break

    cv.destroyAllWindows()


@Provider()
class Camera:
    """
    Camera class for handling camera feeds
    """
    def __init__(self, config: Optional["HoistedConfig"],
                 camera_idx: int = 0,
                 height: int = 480,
                 width: int = 640,
                 fps: float = 30.0,
                 save_video: bool = False,
                 use_gui: bool = False,
                 output_file: Path = Path.cwd().joinpath("camera_output")
                 ):
        """
        Camera class for handling camera feeds

        :param config: The configuration with a hoisted camera config
        :param camera_idx: The index of the camera (used if config is not provided)
        :param height: The height of the camera feed (used if config is not provided)
        :param width: The width of the camera feed (used if config is not provided)
        :param fps: The frames per second of the camera feed (used if config is not provided)
        :param output_file: The output file for saving (used if config is not provided)
        :param save_video: Whether to save the video (used if config is not provided)
        :param use_gui: Whether to use the gui (used if config is not provided)
        """
        ################################################################################################################
        # CONFIGURE
        ################################################################################################################
        #: Config: reference to the configuration
        self.config = config
        #: int: camera index
        self.camera_idx = config.camera_idx if HAS(config) else camera_idx
        #: int: height of the camera feed
        self.height = self.config.height if HAS(config) else height
        #: int: width of the camera feed
        self.width = self.config.width if HAS(config) else width
        #: float: frames per second
        self.fps = self.config.fps if HAS(config) else fps
        #: Path: output file for saving
        self.output_file = self.config.save_location.joinpath(self.config.file_header + f"camera_{self.camera_idx}") \
            if HAS(config) else output_file
        #: bool: whether to use the gui
        self.use_gui = self.config.use_gui if HAS(config) else use_gui
        #: bool: whether to save
        self.save = self.config.save_video if HAS(config) else save_video
        #: Queue: queue of images for display
        self.image_queue = None
        #: Pipe: command pipe
        self.command_pipe = None
        #: Process: camera process
        self.camera = None
        # only set display name if not in gui mode
        self._display_name = None if self.use_gui else f"Camera: {self.camera_idx}"
        self.build()

    def __str__(self) -> str:
        return f"Camera{self.camera_idx}: {self.height}x{self.width} @ {self.fps} fps"

    def __name__(self) -> str:
        return f"Camera{self.camera_idx}"

    @property
    def quarter_period(self) -> float:
        """
        Calculate the quarter period of the camera feed in milliseconds
        """
        return round(1 / self.fps * 1000 / 4)

    def build(self) -> None:
        """
        Build the camera process
        """
        if self.use_gui:
            self.image_queue = Queue(maxsize=1)
        self.command_pipe, command_pipe = Pipe()
        self.camera = Process(name=f"Camera{ self.camera_idx}",
                              target=stream_video,
                              args=(self.camera_idx,
                                    self.height,
                                    self.width,
                                    self.quarter_period,
                                    self._display_name,
                                    self.image_queue,
                                    command_pipe,
                                    self.output_file))

    def start(self) -> None:
        """
        Start the camera process
        """
        self.camera.start()

    def stop(self) -> None:
        """
        Stop the camera process by sending a stop command through the command pipe. This will alert the camera to stop
        and save close the file handle if it was initialized to the save video.
        """
        if self.camera:
            self.command_pipe.send("stop")
            sleep(1)
            self.camera.terminate()
            sleep(1)
            self.camera.join(timeout=5.0)

    def destroy(self) -> None:
        """
        Terminate the camera process
        """
        if self.camera:
            self.camera.terminate()

    def __del__(self):
        if self.camera:
            self.camera.terminate()

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):

        if self.camera:
            self.camera.terminate()


if __name__ == "__main__":
    # Protection from running the module as a script
    pass



from typing import Callable, Tuple, Any
from functools import wraps
from pathlib import Path
import string
import warnings

from ..exceptions import InvalidExtensionWarning, InvalidFilenameError, NotPermittedTypeError


"""
Some functions useful for validation. Most of these functions are parameterized decorators that can be used to
validate function arguments or perform runtime conversion between types that are commensurable but can't be directly
duck-typed.
"""


__all__ = [
    "convert_permitted_types_to_required",
    "report_function_call",
    "validate_extension",
    "validate_filename",
]


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// UTILITIES USED FOR PARAMETERIZATIONS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


def amend_args(arguments: tuple, amendment: Any, pos: int = 0) -> tuple:
    """
    Function amends arguments tuple (~scary tuple mutation~)

    :param arguments: arguments to be amended
    :param amendment: new value of argument
    :param pos: index of argument to be converted
    :returns: amended arguments tuple
    """

    arguments = list(arguments)
    arguments[pos] = amendment
    return tuple(arguments)


def collector(pos: int, key: str, *args, **kwargs) -> Tuple[bool, Any, bool]:
    """
    Function collects the argument to be validated

    :param pos: position of argument to be collected
    :param key: key of argument to be collected
    :param args: arguments for positional collection
    :param kwargs: keyword arguments for keyword collection
    :returns: boolean indicating whether argument was collected, target, and whether the argument was collected
        specifically from positional arguments
    """
    # noinspection PyBroadException
    try:
        if key in kwargs:
            collected = True
            use_args = False
            target = kwargs.get(key)
        elif pos is not None and args[pos] is not None:
            collected = True
            use_args = True
            target = args[pos]
        else:
            raise Exception
    except Exception:  # if any exception, just report a failure to collect
        collected = False
        use_args = None
        target = None
    # noinspection PyUnboundLocalVariable
    return collected, target, use_args


def parameterize(decorator: Callable) -> Callable:
    """
    Function for parameterizing decorators

    :param decorator: A decorator to parameterize

    :returns: A decorator that can be parameterized
    """

    def outer(*args, **kwargs) -> Callable:

        def inner(func: Callable) -> Callable:
            # noinspection PyArgumentList
            return decorator(func, *args, **kwargs)

        return inner

    return outer


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// VALIDATORS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@parameterize
def convert_permitted_types_to_required(function: Callable,
                                        permitted: tuple,
                                        required: Any,
                                        pos: int = 0,
                                        key: str = None,
                                        ) -> Callable:
    """
    Decorator that converts an argument from any of the permitted types to the expected/required type.

    :param function: function to be decorated
    :param permitted: the types permitted by code
    :param required: the type required by code
    :param pos: index of argument to be converted
    :param key: keyword of argument to be converted
    :returns: decorated function

    .. raises:: :class:`NotPermittedTypeError <exceptions.NotPermittedTypeError>`

    .. warning::  The required type must be capable of converting the permitted types using the __call__ magic method.
    """
    @wraps(function)
    def decorator(*args, **kwargs) -> Callable:

        collected, allowed_input, use_args = collector(pos, key, *args, **kwargs)

        if collected:
            if isinstance(allowed_input, permitted):
                allowed_input = required(allowed_input)

            if not isinstance(allowed_input, required):
                raise NotPermittedTypeError(key, pos, permitted)

            if use_args:
                args = amend_args(args, allowed_input, pos)
            else:
                kwargs[key] = allowed_input

        return function(*args, **kwargs)

    return decorator


@parameterize
def validate_extension(function: Callable, required_extension: str, pos: int = 0, key: str = None) \
        -> Callable:  # noqa: U100
    """
    Decorator for validating a required extension on a file path

    :param function: function to be decorated
    :param required_extension: required extension
    :param pos: index of the argument to be validated
    :param key: keyword of the argument to be validated
    :returns: decorated function

    .. raises:: :class:`InvalidExtensionWarning <exceptions.InvalidExtensionWarning>`

    .. note:: This decorator will convert the extension of the file to the required extension if it is not already,
        rather than raising a fatal error.
    """
    @wraps(function)
    def decorator(*args, **kwargs) -> Callable:
        _original_type = type(args[pos])
        if not Path(args[pos]).suffix:
            args = amend_args(args, _original_type("".join([str(args[pos]), required_extension])), pos)
        if Path(args[pos]).suffix != required_extension:
            warnings.warn(InvalidExtensionWarning(key, pos, Path(args[pos]).suffix, required_extension),
                          stacklevel=4)
            args = amend_args(args, _original_type(Path(args[pos]).with_suffix(required_extension)), pos)
        # noinspection PyArgumentList
        return function(*args, **kwargs)
    return decorator


@parameterize
def validate_filename(function: Callable, pos: int = 0, key: str = None) -> Callable:  # noqa: U100
    """
    Decorator for validating filenames adhere to best practices for naming files. Specifically, filenames should only
    contain ascii letters, digits, periods, and underscores. The decorator will validate the entire path, not just
    the filename.

    :param function: function to be decorated
    :param pos: index of the argument to be validated
    :param key: keyword of the argument to be validated
    :returns: decorated function

    .. raises:: :class:`InvalidFilenameError <exceptions.InvalidFilenameError>`

    .. note:: See `here <https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file>1_ for more information
        on file naming best practices for naming files.
    """
    @wraps(function)
    def decorator(*args, **kwargs) -> Callable:

        collected, allowed_input, use_args = collector(pos, key, *args, **kwargs)

        if collected:
            if use_args:
                string_input = str(args[pos])
            else:
                string_input = str(kwargs.get(key))
            string_input = string_input.split("\\")[-1]
            if not set(string_input) <= set(string.ascii_letters + string.digits + "." + "_"):
                raise InvalidFilenameError(key, pos, string_input)

        # noinspection PyArgumentList
        return function(*args, **kwargs)
    return decorator



from typing import Iterable, Any


"""
Backport of pattern matching for python <= 3.10
"""


class PatternMatching:
    def __init__(self, value: Any, comparison_expressions: Iterable[Any]):
        """
        Manual implementation of pattern matching for python < 3.10

        Not the most extensible or robust right now, but she works well for the
        current implementations.

        :param value: value or iterator of values of interest
        :param comparison_expressions: comparison expression or iterator of comparison expressions
        """
        self.value = value
        self.comparison_expressions = self._coerce_iterable(comparison_expressions)

    @staticmethod
    def _coerce_iterable(comparison_expression: Any) -> Iterable[Any]:
        if isinstance(comparison_expression, Iterable):
            return comparison_expression
        else:
            return [comparison_expression, ]  # return as iterable

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):
        return False

    def __call__(self, cases: Any):
        """
        Magic call for comparing some case with some value using some comparison operator

        :param cases: case/s for comparison
        :returns: whether the case/s are matched by the value/s
        """
        for value, comparator, case in zip(self.value, self.comparison_expressions, cases):
            if not comparator(value, case):
                return False
        return True
